# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import contextlib
import ctypes
import errno
import mmap
import os
from pathlib import Path
import re
import struct
import sys
import tempfile
import unittest

from _drgn_util.platform import NORMALIZED_MACHINE_NAME
from drgn import NULL, FaultError, ObjectNotFoundError
from drgn.helpers.linux.device import dev_name
from drgn.helpers.linux.mm import (
    PFN_PHYS,
    PHYS_PFN,
    PageCompound,
    PageHead,
    PageSwapBacked,
    PageTail,
    PageWriteback,
    access_process_vm,
    access_remote_vm,
    cmdline,
    compound_head,
    compound_nr,
    compound_order,
    decode_memory_block_state,
    decode_memory_block_state_value,
    decode_page_flags,
    environ,
    find_vmap_area,
    follow_page,
    follow_pfn,
    follow_phys,
    for_each_memory_block,
    for_each_valid_page_range,
    for_each_vma,
    for_each_vmap_area,
    memory_block_size_bytes,
    page_index,
    page_size,
    page_to_pfn,
    page_to_phys,
    page_to_virt,
    pfn_to_page,
    pfn_to_virt,
    phys_to_page,
    phys_to_virt,
    task_rss,
    task_vsize,
    totalram_pages,
    virt_to_page,
    virt_to_pfn,
    virt_to_phys,
    vm_commit_limit,
    vm_memory_committed,
    vma_find,
    vma_name,
    vmalloc_to_page,
    vmalloc_to_pfn,
)
from drgn.helpers.linux.pid import find_task
from tests.linux_kernel import (
    LinuxKernelTestCase,
    fork_and_stop,
    iter_maps,
    meminfo_field_in_pages,
    mlock,
    prctl_set_vma_anon_name,
    prng32,
    skip_if_highmem,
    skip_if_highpte,
    skip_unless_have_full_mm_support,
    skip_unless_have_memory_hotplug,
    skip_unless_have_test_kmod,
)


class MmTestCase(LinuxKernelTestCase):
    # Returns an mmap.mmap object for a file mapping in /dev/shm, its mapped
    # address, and the pfns backing it.
    @contextlib.contextmanager
    def map_pages(self):
        if not os.path.exists("/proc/self/pagemap"):
            self.skipTest("kernel does not support pagemap")

        pages = 4
        with tempfile.TemporaryFile(dir="/dev/shm") as f:
            f.write(os.urandom(pages * mmap.PAGESIZE))
            f.flush()
            with mmap.mmap(f.fileno(), pages * mmap.PAGESIZE) as map:
                f.close()
                address = ctypes.addressof(ctypes.c_char.from_buffer(map))
                # Make sure the pages are faulted in and stay that way.
                mlock(address, pages * mmap.PAGESIZE)

                with open("/proc/self/pagemap", "rb", buffering=0) as pagemap:
                    pagemap.seek(address // mmap.PAGESIZE * 8)
                    pfns = [
                        entry & ((1 << 54) - 1)
                        for entry in struct.unpack(f"{pages}Q", pagemap.read(pages * 8))
                    ]
                yield map, address, pfns


class TestMm(MmTestCase):
    def test_page_constants(self):
        self.assertEqual(self.prog["PAGE_SIZE"], mmap.PAGESIZE)
        self.assertEqual(1 << self.prog["PAGE_SHIFT"], mmap.PAGESIZE)
        self.assertEqual(~self.prog["PAGE_MASK"] + 1, mmap.PAGESIZE)

    def test_page_index(self):
        with self.map_pages() as (_, _, pfns):
            self.assertEqual(page_index(pfn_to_page(self.prog, pfns[3])), 3)

    def test_page_flag_getters(self):
        with self.map_pages() as (map, _, pfns):
            page = pfn_to_page(self.prog, pfns[0])
            # The page flag getters are generated, so just pick a positive case
            # and a negative case to cover all of them.
            self.assertTrue(PageSwapBacked(page))
            self.assertFalse(PageWriteback(page))

    @skip_unless_have_test_kmod
    def test_PageCompound(self):
        self.assertFalse(PageCompound(self.prog["drgn_test_page"]))
        self.assertTrue(PageCompound(self.prog["drgn_test_compound_page"]))
        self.assertTrue(PageCompound(self.prog["drgn_test_compound_page"] + 1))

    @skip_unless_have_test_kmod
    def test_PageHead(self):
        self.assertFalse(PageHead(self.prog["drgn_test_page"]))
        self.assertTrue(PageHead(self.prog["drgn_test_compound_page"]))
        self.assertFalse(PageHead(self.prog["drgn_test_compound_page"] + 1))

    @skip_unless_have_test_kmod
    def test_PageTail(self):
        self.assertFalse(PageTail(self.prog["drgn_test_page"]))
        self.assertFalse(PageTail(self.prog["drgn_test_compound_page"]))
        self.assertTrue(PageTail(self.prog["drgn_test_compound_page"] + 1))

    @skip_unless_have_test_kmod
    def test_compound_head(self):
        self.assertEqual(
            compound_head(self.prog["drgn_test_page"]), self.prog["drgn_test_page"]
        )
        self.assertEqual(
            compound_head(self.prog["drgn_test_compound_page"]),
            self.prog["drgn_test_compound_page"],
        )
        self.assertEqual(
            compound_head(self.prog["drgn_test_compound_page"] + 1),
            self.prog["drgn_test_compound_page"],
        )

    @skip_unless_have_test_kmod
    def test_compound_order(self):
        self.assertEqual(compound_order(self.prog["drgn_test_page"]), 0)
        self.assertEqual(compound_order(self.prog["drgn_test_compound_page"]), 1)

    @skip_unless_have_test_kmod
    def test_compound_nr(self):
        self.assertEqual(compound_nr(self.prog["drgn_test_page"]), 1)
        self.assertEqual(compound_nr(self.prog["drgn_test_compound_page"]), 2)

    @skip_unless_have_test_kmod
    def test_page_size(self):
        self.assertEqual(page_size(self.prog["drgn_test_page"]), self.prog["PAGE_SIZE"])
        self.assertEqual(
            page_size(self.prog["drgn_test_compound_page"]), 2 * self.prog["PAGE_SIZE"]
        )

    @skip_unless_have_full_mm_support
    def test_decode_page_flags(self):
        with self.map_pages() as (map, _, pfns):
            page = pfn_to_page(self.prog, pfns[0])
            self.assertIn("PG_swapbacked", decode_page_flags(page))

    # In principle, this shouldn't need full MM support, but /proc/kcore is
    # sometimes missing some parts of vmemmap, requiring virtual address
    # translation.
    @skip_unless_have_full_mm_support
    @skip_unless_have_test_kmod
    def test_for_each_valid_page_range(self):
        expected_pfn = self.prog["drgn_test_pfn"].value_()
        found_expected = False
        for start_pfn, end_pfn, mem_map in for_each_valid_page_range(self.prog):
            # We should be able to read all valid pages.
            for page in mem_map[start_pfn:end_pfn]:
                page._refcount.read_()
            if start_pfn <= expected_pfn < end_pfn:
                found_expected = True
                self.assertEqual(
                    mem_map + expected_pfn, self.prog["drgn_test_page"].read_()
                )
        self.assertTrue(found_expected)

    @skip_unless_have_test_kmod
    def test_PFN_PHYS(self):
        self.assertEqual(
            PFN_PHYS(self.prog["drgn_test_pfn"]), self.prog["drgn_test_pa"]
        )

    @skip_unless_have_test_kmod
    def test_PHYS_PFN(self):
        self.assertEqual(
            PHYS_PFN(self.prog["drgn_test_pa"]), self.prog["drgn_test_pfn"]
        )

    @skip_unless_have_test_kmod
    def test_page_to_pfn(self):
        self.assertEqual(
            page_to_pfn(self.prog["drgn_test_page"]), self.prog["drgn_test_pfn"]
        )

    @skip_unless_have_test_kmod
    def test_page_to_phys(self):
        self.assertEqual(
            page_to_phys(self.prog["drgn_test_page"]), self.prog["drgn_test_pa"]
        )

    @skip_unless_have_full_mm_support
    @skip_unless_have_test_kmod
    def test_page_to_virt(self):
        self.assertEqual(
            page_to_virt(self.prog["drgn_test_page"]), self.prog["drgn_test_va"]
        )

    @skip_unless_have_test_kmod
    def test_pfn_to_page(self):
        self.assertEqual(
            pfn_to_page(self.prog["drgn_test_pfn"]), self.prog["drgn_test_page"]
        )

    @skip_unless_have_full_mm_support
    @skip_unless_have_test_kmod
    def test_pfn_to_virt(self):
        self.assertEqual(
            pfn_to_virt(self.prog["drgn_test_pfn"]), self.prog["drgn_test_va"]
        )

    @skip_unless_have_test_kmod
    def test_phys_to_page(self):
        self.assertEqual(
            phys_to_page(self.prog["drgn_test_pa"]), self.prog["drgn_test_page"]
        )

    @skip_unless_have_full_mm_support
    @skip_unless_have_test_kmod
    def test_phys_to_virt(self):
        self.assertEqual(
            phys_to_virt(self.prog["drgn_test_pa"]), self.prog["drgn_test_va"]
        )

    @skip_unless_have_full_mm_support
    @skip_unless_have_test_kmod
    def test_virt_to_page(self):
        self.assertEqual(
            virt_to_page(self.prog["drgn_test_va"]), self.prog["drgn_test_page"]
        )

    @skip_unless_have_full_mm_support
    @skip_unless_have_test_kmod
    def test_virt_to_pfn(self):
        self.assertEqual(
            virt_to_pfn(self.prog["drgn_test_va"]), self.prog["drgn_test_pfn"]
        )

    @skip_unless_have_full_mm_support
    @skip_unless_have_test_kmod
    def test_virt_to_phys(self):
        self.assertEqual(
            virt_to_phys(self.prog["drgn_test_va"]), self.prog["drgn_test_pa"]
        )

    @skip_unless_have_test_kmod
    def test_read_physical(self):
        expected = bytearray()
        for x in prng32("PAGE"):
            expected.extend(x.to_bytes(4, sys.byteorder))
            if len(expected) >= mmap.PAGESIZE:
                break
        self.assertEqual(
            self.prog.read(self.prog["drgn_test_pa"], mmap.PAGESIZE, True), expected
        )

    @skip_unless_have_full_mm_support
    @skip_unless_have_test_kmod
    def test_follow_phys(self):
        self.assertEqual(
            follow_phys(self.prog["init_mm"].address_of_(), self.prog["drgn_test_va"]),
            self.prog["drgn_test_pa"],
        )

    @skip_unless_have_full_mm_support
    @skip_unless_have_test_kmod
    def test_follow_page(self):
        self.assertEqual(
            follow_page(self.prog["init_mm"].address_of_(), self.prog["drgn_test_va"]),
            self.prog["drgn_test_page"],
        )

    @skip_unless_have_full_mm_support
    @skip_if_highpte
    def test_follow_pfn(self):
        task = find_task(self.prog, os.getpid())
        with self.map_pages() as (map, address, pfns):
            self.assertEqual(follow_pfn(task.mm, address), pfns[0])

    @skip_unless_have_full_mm_support
    @skip_unless_have_test_kmod
    def test_follow_pfn_init_mm(self):
        self.assertEqual(
            follow_pfn(self.prog["init_mm"].address_of_(), self.prog["drgn_test_va"]),
            self.prog["drgn_test_pfn"],
        )

    @skip_unless_have_full_mm_support
    @skip_unless_have_test_kmod
    def test_vmalloc_to_page(self):
        self.assertEqual(
            vmalloc_to_page(self.prog["drgn_test_vmalloc_va"]),
            self.prog["drgn_test_vmalloc_page"],
        )

    @skip_unless_have_full_mm_support
    @skip_unless_have_test_kmod
    def test_vmalloc_to_pfn(self):
        self.assertEqual(
            vmalloc_to_pfn(self.prog["drgn_test_vmalloc_va"]),
            self.prog["drgn_test_vmalloc_pfn"],
        )

    @skip_unless_have_test_kmod
    def test_find_vmap_area(self):
        self.assertEqual(
            find_vmap_area(
                self.prog, self.prog["drgn_test_vmalloc_va"] + 1234
            ).va_start.value_(),
            self.prog["drgn_test_vmalloc_va"].value_(),
        )

        with self.subTest("non-vmap address"):
            self.assertIdentical(
                find_vmap_area(self.prog, self.prog["drgn_test_va"]),
                NULL(self.prog, "struct vmap_area *"),
            )

    @skip_unless_have_test_kmod
    def test_for_each_vmap_area(self):
        self.assertTrue(
            any(
                va.va_start.value_() == self.prog["drgn_test_vmalloc_va"].value_()
                for va in for_each_vmap_area(self.prog)
            )
        )

    @skip_unless_have_full_mm_support
    @skip_if_highmem
    def test_access_process_vm(self):
        task = find_task(self.prog, os.getpid())
        data = b"hello, world"
        buf = ctypes.create_string_buffer(data)
        address = ctypes.addressof(buf)
        self.assertEqual(access_process_vm(task, address, len(data)), data)
        self.assertEqual(access_remote_vm(task.mm, address, len(data)), data)
        self.assertRaises(FaultError, access_process_vm, task, 0, 8)

    @skip_unless_have_full_mm_support
    @skip_if_highmem
    def test_access_process_vm_big(self):
        task = find_task(self.prog, os.getpid())
        # 32M = 2**(log2(16K) + log2(16K / 8)), so 32MB + 1 is enough so that
        # even with 16KB pages on a 64-bit architecture, we're guaranteed to
        # read across two second-level page table entries. (The same for 64KB
        # pages is 512MB + 1, which is too big to test reliably in vmtest.)
        max_size = 32 * 1024 * 1024 + 1
        expected = os.getrandom(max_size)
        while len(expected) < max_size:
            expected += os.getrandom(max_size - len(expected))
        address = ctypes.memmove(expected, expected, 0)
        expected = memoryview(expected)
        for size in (1, 65537, max_size):
            with self.subTest(size=size):
                self.assertEqual(
                    access_process_vm(task, address, size), expected[:size]
                )

    @skip_unless_have_full_mm_support
    def test_access_remote_vm_init_mm(self):
        data = self.prog["UTS_RELEASE"].string_()
        self.assertEqual(
            access_remote_vm(
                self.prog["init_mm"].address_of_(),
                self.prog["init_uts_ns"].name.release + 0,
                len(data),
            ),
            data,
        )

    @unittest.skipUnless(NORMALIZED_MACHINE_NAME == "x86_64", "machine is not x86_64")
    def test_non_canonical_x86_64(self):
        task = find_task(self.prog, os.getpid())
        data = b"hello, world"
        buf = ctypes.create_string_buffer(data)
        address = ctypes.addressof(buf)
        self.assertRaises(
            FaultError, access_process_vm, task, address ^ (1 << 63), len(data)
        )

    @skip_unless_have_full_mm_support
    @skip_if_highmem
    def test_cmdline(self):
        with open("/proc/self/cmdline", "rb") as f:
            proc_cmdline = f.read().split(b"\0")[:-1]
        task = find_task(self.prog, os.getpid())
        self.assertEqual(cmdline(task), proc_cmdline)

    def test_cmdline_kernel_thread(self):
        self.assertIsNone(cmdline(find_task(self.prog, 2)))

    @skip_unless_have_full_mm_support
    @skip_if_highmem
    def test_environ(self):
        with open("/proc/self/environ", "rb") as f:
            proc_environ = f.read().split(b"\0")[:-1]
        task = find_task(self.prog, os.getpid())
        self.assertEqual(environ(task), proc_environ)

    def test_environ_kernel_thread(self):
        self.assertIsNone(environ(find_task(self.prog, 2)))

    def test_vma_find(self):
        with fork_and_stop() as pid:
            mm = find_task(self.prog, pid).mm

            prev_end = 0
            for map in iter_maps(pid):
                # Gate VMAs are not included in vma_find().
                if map.is_gate():
                    continue

                if map.start != prev_end:
                    self.assertIdentical(
                        vma_find(mm, prev_end),
                        NULL(self.prog, "struct vm_area_struct *"),
                    )
                    self.assertIdentical(
                        vma_find(mm, map.start - 1),
                        NULL(self.prog, "struct vm_area_struct *"),
                    )
                vma = vma_find(mm, map.start)
                self.assertEqual((map.start, map.end), (vma.vm_start, vma.vm_end))

                vma = vma_find(mm, (map.start + map.end) // 2)
                self.assertEqual((map.start, map.end), (vma.vm_start, vma.vm_end))

                vma = vma_find(mm, map.end - 1)
                self.assertEqual((map.start, map.end), (vma.vm_start, vma.vm_end))

                prev_end = map.end

            self.assertIdentical(
                vma_find(mm, prev_end), NULL(self.prog, "struct vm_area_struct *")
            )

    def test_vma_name(self):
        with mmap.mmap(-1, mmap.PAGESIZE, mmap.MAP_PRIVATE) as private_map, mmap.mmap(
            -1, mmap.PAGESIZE, mmap.MAP_SHARED
        ) as shared_map:
            # Test VMA names if the kernel supports it.
            try:
                prctl_set_vma_anon_name(
                    ctypes.addressof(ctypes.c_char.from_buffer(private_map)),
                    mmap.PAGESIZE,
                    "testprivate",
                )
            except OSError as e:
                # PR_SET_VMA_ANON_NAME is only supported since Linux 5.17, and
                # only if CONFIG_ANON_VMA_NAME=y. Otherwise, it returns EINVAL.
                if e.errno != errno.EINVAL:
                    raise
            try:
                prctl_set_vma_anon_name(
                    ctypes.addressof(ctypes.c_char.from_buffer(shared_map)),
                    mmap.PAGESIZE,
                    "testshared",
                )
            except OSError as e:
                # Unsupported VMAs return EBADF, and anonymous shared memory is
                # only supported since Linux 6.2.
                if e.errno != errno.EINVAL and e.errno != errno.EBADF:
                    raise

            mm = find_task(self.prog, os.getpid()).mm.read_()
            tested_file_path = False
            for map in iter_maps():
                if map.path.startswith("["):
                    vma = vma_find(mm, map.start)
                    if vma:
                        with self.subTest(vma=map.path):
                            self.assertEqual(vma_name(vma), os.fsencode(map.path))
                elif not tested_file_path and map.path.startswith("/"):
                    vma = vma_find(mm, map.start)
                    with self.subTest("file"):
                        self.assertEqual(vma_name(vma), os.fsencode(map.path))
                    tested_file_path = True

    def test_for_each_vma(self):
        with fork_and_stop() as pid:
            self.assertEqual(
                [
                    (vma.vm_start, vma.vm_end)
                    for vma in for_each_vma(find_task(self.prog, pid).mm)
                ],
                [
                    (map.start, map.end)
                    for map in iter_maps(pid)
                    # Gate VMAs are not included in for_each_vma().
                    if not map.is_gate()
                ],
            )

    def test_totalram_pages(self):
        self.assertEqual(totalram_pages(self.prog), meminfo_field_in_pages("MemTotal"))

    def test_vm_commit_limit(self):
        overcommit_kbytes_path = Path("/proc/sys/vm/overcommit_kbytes")
        orig_overcommit_kbytes = int(overcommit_kbytes_path.read_text())
        try:
            for i in range(2):
                if i == 1:
                    if orig_overcommit_kbytes == 0:
                        overcommit_kbytes = 1024 * 1024 * 1024
                    else:
                        overcommit_kbytes = 0
                    overcommit_kbytes_path.write_text(str(overcommit_kbytes))
                self.assertEqual(
                    vm_commit_limit(self.prog), meminfo_field_in_pages("CommitLimit")
                )
        finally:
            overcommit_kbytes_path.write_text(str(orig_overcommit_kbytes))

    def test_vm_memory_committed(self):
        self.assertAlmostEqual(
            vm_memory_committed(self.prog),
            meminfo_field_in_pages("Committed_AS"),
            delta=1024 * 1024 * 1024,
        )

    def test_task_rss(self):
        with fork_and_stop() as pid:
            task = find_task(self.prog, pid)
            rss_info = task_rss(self.prog, task)

            page_size = self.prog["PAGE_SIZE"].value_()
            # Get the relevant RSS counters, converting from kB to pages.
            stats = {
                key: int(value) * 1024 // page_size
                for key, value in re.findall(
                    r"^(VmRSS|RssAnon|RssFile|RssShmem|VmSwap):\s*([0-9]+)",
                    Path(f"/proc/{pid}/status").read_text(),
                    flags=re.MULTILINE,
                )
            }

            # Before Linux kernel commit 82241a83cd15 ("mm: fix the inaccurate
            # memory statistics issue for users") (in v6.16), the RSS counters
            # in /proc/pid/meminfo are approximate due to batching, but the
            # helpers are exact.
            if hasattr(task, "rss_stat"):
                # Before Linux kernel commit f1a7941243c10 ("mm: convert mm's
                # rss stats into percpu_counter") (in v6.2), there is a
                # per-thread counter that only gets synced to the main counter
                # every TASK_RSS_EVENTS_THRESH (64) page faults. Each fault can
                # map in multiple pages based on fault_around_bytes. So, the
                # maximum error is nr_threads * 64 * (fault_around_bytes / PAGE_SIZE).
                delta = (
                    len(os.listdir(f"/proc/{pid}/task"))
                    * 64
                    * (self.prog["fault_around_bytes"].value_() // page_size)
                )
            else:
                # Between that and Linux kernel commit 82241a83cd15 ("mm: fix
                # the inaccurate memory statistics issue for users") (in
                # v6.16), the kernel code uses percpu_counter_read_positive(),
                # so the maximum error is nr_cpus * percpu_counter_batch.
                try:
                    percpu_counter_batch = self.prog["percpu_counter_batch"].value_()
                except ObjectNotFoundError:
                    percpu_counter_batch = 32
                delta = percpu_counter_batch * os.cpu_count()

            self.assertAlmostEqual(rss_info.file, stats["RssFile"], delta=delta)
            self.assertAlmostEqual(rss_info.anon, stats["RssAnon"], delta=delta)
            self.assertAlmostEqual(
                rss_info.shmem, stats.get("RssShmem", 0), delta=delta
            )
            self.assertAlmostEqual(rss_info.swap, stats["VmSwap"], delta=delta)
            # VmRSS is the sum of three counters, so it has triple the error
            # margin.
            self.assertAlmostEqual(rss_info.total, stats["VmRSS"], delta=delta * 3)

    def test_task_vsize(self):
        with fork_and_stop() as pid:
            task = find_task(self.prog, pid)
            vsize = task_vsize(task)
            text = Path(f"/proc/{pid}/status").read_text()
            value = re.findall(r"^VmSize:\s*([0-9]+)", text, flags=re.MULTILINE)
            if value:
                self.assertEqual(vsize, int(value[0]) * 1024)

    @skip_unless_have_memory_hotplug
    def test_for_each_memory_block(self):
        self.assertCountEqual(
            [
                dev_name(mem.dev.address_of_())
                for mem in for_each_memory_block(self.prog)
            ],
            os.listdir(b"/sys/bus/memory/devices"),
        )

    @skip_unless_have_memory_hotplug
    def test_decode_memory_block_state(self):
        mem = next(iter(for_each_memory_block(self.prog)))
        self.assertEqual(
            decode_memory_block_state(mem)
            .replace("MEM_", "")
            .lower()
            .replace("_", "-"),
            (
                Path("/sys/bus/memory/devices")
                / os.fsdecode(dev_name(mem.dev.address_of_()))
                / "state"
            )
            .read_text()
            .strip(),
        )

    @skip_unless_have_test_kmod
    def test_decode_memory_block_state_value(self):
        for name in (
            "MEM_ONLINE",
            "MEM_GOING_OFFLINE",
            "MEM_OFFLINE",
            "MEM_GOING_ONLINE",
            "MEM_CANCEL_ONLINE",
            "MEM_CANCEL_OFFLINE",
            "MEM_PREPARE_ONLINE",
            "MEM_FINISH_OFFLINE",
        ):
            with self.subTest(state=name):
                try:
                    state = self.prog["drgn_test_" + name]
                except ObjectNotFoundError:
                    self.skipTest(f"{name} is not defined")
                self.assertEqual(decode_memory_block_state_value(state), name)

    @skip_unless_have_memory_hotplug
    def test_memory_block_size_bytes(self):
        self.assertEqual(
            memory_block_size_bytes(self.prog),
            int(Path("/sys/devices/system/memory/block_size_bytes").read_text(), 16),
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import unittest

import drgn
from drgn import (
    Architecture,
    Object,
    Platform,
    Program,
    search_memory,
    search_memory_regex,
    search_memory_u32,
    search_memory_u64,
    search_memory_word,
)
from tests import (
    MOCK_32BIT_BIG_ENDIAN_PLATFORM,
    MOCK_32BIT_PLATFORM,
    MOCK_BIG_ENDIAN_PLATFORM,
    MOCK_PLATFORM,
    IntWrapper,
    MockMemorySegment,
    TestCase,
    add_mock_memory_segments,
    with_default_prog,
)


def mock_search_memory_program(*segments, platform=MOCK_PLATFORM):
    prog = Program(platform)
    add_mock_memory_segments(prog, segments)
    return prog


class TestSearchMemoryBytes(TestCase):
    def test_one_match(self):
        prog = mock_search_memory_program(MockMemorySegment(b"abcdef", 0x1000))
        self.assertEqual(list(prog.search_memory(b"cde")), [0x1002])

    def test_haystack_embedded_null(self):
        prog = mock_search_memory_program(MockMemorySegment(b"a\0bcdef\0g", 0x1000))
        self.assertEqual(list(prog.search_memory(b"cde")), [0x1003])

    def test_needle_embedded_null(self):
        prog = mock_search_memory_program(MockMemorySegment(b"ab\0c\0d", 0x1000))
        self.assertEqual(list(prog.search_memory(b"b\0c")), [0x1001])

    def test_multiple_matches(self):
        prog = mock_search_memory_program(MockMemorySegment(b"abcd abcz", 0x1000))
        self.assertEqual(list(prog.search_memory(b"abc")), [0x1000, 0x1005])

    def test_non_overlapping(self):
        prog = mock_search_memory_program(MockMemorySegment(b"aaaa", 0x1000))
        self.assertEqual(list(prog.search_memory(b"aa")), [0x1000, 0x1002])

    def test_address_range(self):
        prog = mock_search_memory_program(MockMemorySegment(b"aaaa", 0x1000))

        self.assertEqual(
            list(prog.search_memory(b"aa").set_address_range(0x1000, 0x1003)),
            [0x1000, 0x1002],
        )

        self.assertEqual(
            list(prog.search_memory(b"aa").set_address_range(0x1001)), [0x1001]
        )

        self.assertEqual(
            list(prog.search_memory(b"aa").set_address_range(0x1000, 0x1000)), []
        )

    def test_address_range_reuse(self):
        prog = mock_search_memory_program(MockMemorySegment(b"aaaa", 0x1000))

        it = prog.search_memory(b"aa")

        it.set_address_range(0x1000, 0x1003)
        self.assertEqual(list(it), [0x1000, 0x1002])

        it.set_address_range(0x1001)
        self.assertEqual(list(it), [0x1001])

        it.set_address_range(0x1000, 0x1000)
        self.assertEqual(list(it), [])

    def test_invalid_address_range(self):
        self.assertRaises(
            ValueError,
            mock_search_memory_program().search_memory(b"foo").set_address_range,
            0x1000,
            0xFFF,
        )

    def test_cross_boundary(self):
        prog = mock_search_memory_program(MockMemorySegment(b"foo bar", 0x3FFFFFFA))
        self.assertEqual(list(prog.search_memory(b"bar")), [0x3FFFFFFE])

    def test_gap(self):
        prog = mock_search_memory_program(
            MockMemorySegment(b"abcdefgh", 0x1000),
            MockMemorySegment(b"zcdevwxy", 0x1104),
        )
        self.assertEqual(list(prog.search_memory(b"cde")), [0x1002, 0x1105])

    def test_at_max_address(self):
        prog = mock_search_memory_program(
            MockMemorySegment(b"aaaa", 0xFFFFFFFFFFFFFFFC)
        )
        self.assertEqual(
            list(prog.search_memory(b"aa")), [0xFFFFFFFFFFFFFFFC, 0xFFFFFFFFFFFFFFFE]
        )

    def test_one_byte_at_max_address(self):
        prog = mock_search_memory_program(MockMemorySegment(b"a", 0xFFFFFFFFFFFFFFFF))
        self.assertEqual(list(prog.search_memory(b"a")), [0xFFFFFFFFFFFFFFFF])

    def test_at_max_address_32_bit(self):
        prog = mock_search_memory_program(
            MockMemorySegment(b"aaaa", 0xFFFFFFFC),
            platform=MOCK_32BIT_PLATFORM,
        )
        self.assertEqual(list(prog.search_memory(b"aa")), [0xFFFFFFFC, 0xFFFFFFFE])

    def test_one_byte_at_max_address_32_bit(self):
        prog = mock_search_memory_program(
            MockMemorySegment(b"a", 0xFFFFFFFF),
            platform=MOCK_32BIT_PLATFORM,
        )
        self.assertEqual(list(prog.search_memory(b"a")), [0xFFFFFFFF])

    def test_alignment(self):
        prog = mock_search_memory_program(
            MockMemorySegment(b"aaaaaaaa", 0xFFC),
        )
        self.assertEqual(
            list(prog.search_memory(b"aa", alignment=8)),
            [0x1000],
        )

    def test_alignment_at_max_address(self):
        prog = mock_search_memory_program(
            MockMemorySegment(b"a" * 12, 0xFFFFFFFFFFFFFFF4),
        )
        self.assertEqual(
            list(prog.search_memory(b"aa", alignment=8)),
            [0xFFFFFFFFFFFFFFF8],
        )

    def test_u32(self):
        prog = mock_search_memory_program(
            MockMemorySegment(
                b"".join([i.to_bytes(4, "little") for i in range(1, 4)]), 0x1000
            ),
        )
        self.assertEqual(
            list(prog.search_memory((2).to_bytes(4, "little"), alignment=4)), [0x1004]
        )

    def test_u32_big_endian(self):
        prog = mock_search_memory_program(
            MockMemorySegment(
                b"".join([i.to_bytes(4, "big") for i in range(1, 4)]), 0x1000
            ),
            platform=MOCK_BIG_ENDIAN_PLATFORM,
        )
        self.assertEqual(
            list(prog.search_memory((2).to_bytes(4, "big"), alignment=4)), [0x1004]
        )

    def test_u64(self):
        prog = mock_search_memory_program(
            MockMemorySegment(
                b"".join([i.to_bytes(8, "little") for i in range(1, 4)]), 0x1000
            ),
        )
        self.assertEqual(
            list(prog.search_memory((2).to_bytes(8, "little"), alignment=8)), [0x1008]
        )

    def test_u64_big_endian(self):
        prog = mock_search_memory_program(
            MockMemorySegment(
                b"".join([i.to_bytes(8, "big") for i in range(1, 4)]), 0x1000
            ),
            platform=MOCK_BIG_ENDIAN_PLATFORM,
        )
        self.assertEqual(
            list(prog.search_memory((2).to_bytes(8, "big"), alignment=8)), [0x1008]
        )

    def test_no_segments(self):
        prog = mock_search_memory_program()
        self.assertEqual(list(prog.search_memory(b"foo")), [])

    def test_empty(self):
        with self.assertRaisesRegex(ValueError, "needle cannot be empty"):
            next(mock_search_memory_program().search_memory(b""))

    def test_invalid_alignment(self):
        with self.assertRaisesRegex(ValueError, "alignment must be power of 2"):
            next(mock_search_memory_program().search_memory(b"foo", alignment=3))

    def test_invalid_needle_type(self):
        self.assertRaises(TypeError, mock_search_memory_program().search_memory, {})

    def test_default_program(self):
        with with_default_prog(
            mock_search_memory_program(MockMemorySegment(b"abcdef", 0x1000))
        ):
            self.assertEqual(list(search_memory(b"cde")), [0x1002])


class TestSearchMemoryStr(TestCase):
    def test_valid_utf8(self):
        prog = mock_search_memory_program(MockMemorySegment(b"abcdef", 0x1000))
        self.assertEqual(list(prog.search_memory("cde")), [0x1002])

    def test_haystack_embedded_null(self):
        prog = mock_search_memory_program(MockMemorySegment(b"a\0bcdef\0g", 0x1000))
        self.assertEqual(list(prog.search_memory("cde")), [0x1003])

    def test_needle_embedded_null(self):
        prog = mock_search_memory_program(MockMemorySegment(b"ab\0c\0d", 0x1000))
        self.assertEqual(list(prog.search_memory("b\0c")), [0x1001])

    def test_invalid_utf8(self):
        prog = mock_search_memory_program(
            MockMemorySegment(b"\xc3\x28abcdef\xa0\xa1", 0x1000)
        )
        self.assertEqual(list(prog.search_memory("cde")), [0x1004])

    def test_empty(self):
        with self.assertRaisesRegex(ValueError, "empty"):
            next(mock_search_memory_program().search_memory(""))

    def test_alignment(self):
        prog = mock_search_memory_program(
            MockMemorySegment(b"aaaaaaaa", 0xFFC),
        )
        self.assertEqual(
            list(prog.search_memory("aa", alignment=8)),
            [0x1000],
        )

    def test_invalid_alignment(self):
        with self.assertRaisesRegex(ValueError, "alignment"):
            next(mock_search_memory_program().search_memory("foo", alignment=3))

    def test_default_program(self):
        with with_default_prog(
            mock_search_memory_program(MockMemorySegment(b"abcdef", 0x1000))
        ):
            self.assertEqual(list(search_memory("cde")), [0x1002])


class TestSearchMemoryInt(TestCase):
    def test_64_bit(self):
        prog = mock_search_memory_program(
            MockMemorySegment(
                b"\x02\x00\x00\x00\x02\x00\x00\x00\x02\x00\x00\x00\x00\x00\x00\x00",
                0x1000,
            ),
        )
        self.assertEqual(list(prog.search_memory(2)), [0x1008])

    def test_64_bit_big_endian(self):
        prog = mock_search_memory_program(
            MockMemorySegment(
                b"\x00\x00\x00\x02\x00\x00\x00\x02\x00\x00\x00\x00\x00\x00\x00\x02",
                0x1000,
            ),
            platform=MOCK_BIG_ENDIAN_PLATFORM,
        )
        self.assertEqual(list(prog.search_memory(2)), [0x1008])

    def test_32_bit(self):
        prog = mock_search_memory_program(
            MockMemorySegment(
                b"\x02\x00\x00\x00\x02\x00\x00\x00\x02\x00\x00\x00\x00\x00\x00\x00",
                0x1000,
            ),
            platform=MOCK_32BIT_PLATFORM,
        )
        self.assertEqual(list(prog.search_memory(2)), [0x1000, 0x1004, 0x1008])

    def test_32_bit_big_endian(self):
        prog = mock_search_memory_program(
            MockMemorySegment(
                b"\x00\x00\x00\x02\x00\x00\x00\x02\x00\x00\x00\x00\x00\x00\x00\x02",
                0x1000,
            ),
            platform=MOCK_32BIT_BIG_ENDIAN_PLATFORM,
        )
        self.assertEqual(list(prog.search_memory(2)), [0x1000, 0x1004, 0x100C])

    def test_index(self):
        prog = mock_search_memory_program(
            MockMemorySegment(
                b"".join([i.to_bytes(8, "little") for i in range(1, 4)]), 0x1000
            ),
        )
        self.assertEqual(list(prog.search_memory(IntWrapper(2))), [0x1008])

    def test_invalid_alignment(self):
        with self.assertRaisesRegex(TypeError, "alignment"):
            next(mock_search_memory_program().search_memory(0, alignment=3))

    def test_default_program(self):
        with with_default_prog(
            mock_search_memory_program(
                MockMemorySegment(
                    b"\x02\x00\x00\x00\x02\x00\x00\x00\x02\x00\x00\x00\x00\x00\x00\x00",
                    0x1000,
                ),
            )
        ):
            self.assertEqual(list(search_memory(2)), [0x1008])


# Note that these tests use actual architectures so that they have alignment
# information.
class TestSearchMemoryObject(TestCase):
    def test_u32(self):
        prog = mock_search_memory_program(
            MockMemorySegment(
                b"".join([i.to_bytes(4, "little") for i in range(1, 4)]), 0x1000
            ),
            platform=Platform(Architecture.X86_64),
        )
        self.assertEqual(list(prog.search_memory(Object(prog, "int", 2))), [0x1004])

    def test_u32_big_endian(self):
        prog = mock_search_memory_program(
            MockMemorySegment(
                b"".join([i.to_bytes(4, "big") for i in range(1, 4)]), 0x1000
            ),
            platform=Platform(Architecture.S390X),
        )
        self.assertEqual(list(prog.search_memory(Object(prog, "int", 2))), [0x1004])

    def test_u64(self):
        prog = mock_search_memory_program(
            MockMemorySegment(
                b"".join([i.to_bytes(8, "little") for i in range(1, 4)]), 0x1000
            ),
            platform=Platform(Architecture.X86_64),
        )
        self.assertEqual(list(prog.search_memory(Object(prog, "long", 2))), [0x1008])

    def test_u64_big_endian(self):
        prog = mock_search_memory_program(
            MockMemorySegment(
                b"".join([i.to_bytes(8, "big") for i in range(1, 4)]), 0x1000
            ),
            platform=Platform(Architecture.S390X),
        )
        self.assertEqual(list(prog.search_memory(Object(prog, "long", 2))), [0x1008])

    def test_small_object(self):
        prog = mock_search_memory_program(
            MockMemorySegment(b"\x01\x02\x03\x04", 0x1000),
            platform=Platform(Architecture.X86_64),
        )
        self.assertEqual(list(prog.search_memory(Object(prog, "char", 2))), [0x1001])

    def test_large_object(self):
        prog = mock_search_memory_program(
            MockMemorySegment(
                b"".join([i.to_bytes(8, "little") for i in range(1, 9)]), 0x1000
            ),
            platform=Platform(Architecture.X86_64),
        )
        self.assertEqual(
            list(prog.search_memory(Object(prog, "long [4]", [2, 3, 4, 5]))), [0x1008]
        )

    def test_alignment(self):
        prog = mock_search_memory_program(
            MockMemorySegment(b"\x02\x00\x00\x00" * 7, 0xFFC),
            platform=Platform(Architecture.X86_64),
        )
        self.assertEqual(
            list(prog.search_memory(Object(prog, "long long [3]", [0x200000002] * 3))),
            [0x1000],
        )

        prog = mock_search_memory_program(
            MockMemorySegment(b"\x02\x00\x00\x00" * 7, 0xFFC),
            platform=Platform(Architecture.I386),
        )
        self.assertEqual(
            list(prog.search_memory(Object(prog, "long long [3]", [0x200000002] * 3))),
            [0xFFC],
        )

    def test_empty(self):
        prog = mock_search_memory_program(platform=Platform(Architecture.X86_64))
        with self.assertRaisesRegex(ValueError, "size 0"):
            prog.search_memory(Object(prog, "int [0]", address=0x1000))

    def test_invalid_alignment(self):
        prog = mock_search_memory_program(platform=Platform(Architecture.X86_64))
        with self.assertRaisesRegex(TypeError, "alignment"):
            next(prog.search_memory(Object(prog, "int", 0), alignment=3))

    def test_default_program(self):
        prog = mock_search_memory_program(
            MockMemorySegment(
                b"".join([i.to_bytes(4, "little") for i in range(1, 4)]), 0x1000
            ),
            platform=Platform(Architecture.X86_64),
        )
        self.assertEqual(list(search_memory(Object(prog, "int", 2))), [0x1004])

    def test_out_of_range(self):
        with self.assertRaises(OverflowError):
            next(
                mock_search_memory_program(platform=MOCK_32BIT_PLATFORM).search_memory(
                    2**32
                )
            )
        self.assertEqual(list(mock_search_memory_program().search_memory(2**32)), [])
        with self.assertRaises(OverflowError):
            next(mock_search_memory_program().search_memory(2**64))


class TestSearchMemoryUint(TestCase):
    def test_one_u16(self):
        prog = mock_search_memory_program(
            MockMemorySegment(
                b"".join([i.to_bytes(2, "little") for i in range(1, 4)]), 0x1000
            ),
        )
        self.assertEqual(list(prog.search_memory_u16(2)), [(0x1002, 2)])

    def test_one_u16_big_endian(self):
        prog = mock_search_memory_program(
            MockMemorySegment(
                b"".join([i.to_bytes(2, "big") for i in range(1, 4)]), 0x1000
            ),
            platform=MOCK_BIG_ENDIAN_PLATFORM,
        )
        self.assertEqual(list(prog.search_memory_u16(2)), [(0x1002, 2)])

    def test_one_u32(self):
        prog = mock_search_memory_program(
            MockMemorySegment(
                b"".join([i.to_bytes(4, "little") for i in range(1, 4)]), 0x1000
            ),
        )
        self.assertEqual(list(prog.search_memory_u32(2)), [(0x1004, 2)])

    def test_one_u32_big_endian(self):
        prog = mock_search_memory_program(
            MockMemorySegment(
                b"".join([i.to_bytes(4, "big") for i in range(1, 4)]), 0x1000
            ),
            platform=MOCK_BIG_ENDIAN_PLATFORM,
        )
        self.assertEqual(list(prog.search_memory_u32(2)), [(0x1004, 2)])

    def test_one_u64(self):
        prog = mock_search_memory_program(
            MockMemorySegment(
                b"".join([i.to_bytes(8, "little") for i in range(1, 4)]), 0x1000
            ),
        )
        self.assertEqual(list(prog.search_memory_u64(2)), [(0x1008, 2)])

    def test_one_u64_big_endian(self):
        prog = mock_search_memory_program(
            MockMemorySegment(
                b"".join([i.to_bytes(8, "big") for i in range(1, 4)]), 0x1000
            ),
            platform=MOCK_BIG_ENDIAN_PLATFORM,
        )
        self.assertEqual(list(prog.search_memory_u64(2)), [(0x1008, 2)])

    def test_word_64_bit(self):
        prog = mock_search_memory_program(
            MockMemorySegment(
                b"\x02\x00\x00\x00\x02\x00\x00\x00\x02\x00\x00\x00\x00\x00\x00\x00",
                0x1000,
            ),
        )
        self.assertEqual(list(prog.search_memory_word(2)), [(0x1008, 2)])

    def test_word_32_bit(self):
        prog = mock_search_memory_program(
            MockMemorySegment(
                b"\x02\x00\x00\x00\x02\x00\x00\x00\x02\x00\x00\x00\x00\x00\x00\x00",
                0x1000,
            ),
            platform=MOCK_32BIT_PLATFORM,
        )
        self.assertEqual(
            list(prog.search_memory_word(2)), [(0x1000, 2), (0x1004, 2), (0x1008, 2)]
        )

    def test_multiple_u16(self):
        prog = mock_search_memory_program(
            MockMemorySegment(
                b"".join([i.to_bytes(2, "little") for i in range(1, 5)]), 0x1000
            ),
        )
        self.assertEqual(list(prog.search_memory_u16(2, 4)), [(0x1002, 2), (0x1006, 4)])

    def test_multiple_u16_big_endian(self):
        prog = mock_search_memory_program(
            MockMemorySegment(
                b"".join([i.to_bytes(2, "big") for i in range(1, 5)]), 0x1000
            ),
            platform=MOCK_BIG_ENDIAN_PLATFORM,
        )
        self.assertEqual(list(prog.search_memory_u16(2, 4)), [(0x1002, 2), (0x1006, 4)])

    def test_multiple_u32(self):
        prog = mock_search_memory_program(
            MockMemorySegment(
                b"".join([i.to_bytes(4, "little") for i in range(1, 5)]), 0x1000
            ),
        )
        self.assertEqual(list(prog.search_memory_u32(2, 4)), [(0x1004, 2), (0x100C, 4)])

    def test_multiple_u32_big_endian(self):
        prog = mock_search_memory_program(
            MockMemorySegment(
                b"".join([i.to_bytes(4, "big") for i in range(1, 5)]), 0x1000
            ),
            platform=MOCK_BIG_ENDIAN_PLATFORM,
        )
        self.assertEqual(list(prog.search_memory_u32(2, 4)), [(0x1004, 2), (0x100C, 4)])

    def test_multiple_u64(self):
        prog = mock_search_memory_program(
            MockMemorySegment(
                b"".join([i.to_bytes(8, "little") for i in range(1, 5)]), 0x1000
            ),
        )
        self.assertEqual(list(prog.search_memory_u64(2, 4)), [(0x1008, 2), (0x1018, 4)])

    def test_multiple_u64_big_endian(self):
        prog = mock_search_memory_program(
            MockMemorySegment(
                b"".join([i.to_bytes(8, "big") for i in range(1, 5)]), 0x1000
            ),
            platform=MOCK_BIG_ENDIAN_PLATFORM,
        )
        self.assertEqual(list(prog.search_memory_u64(2, 4)), [(0x1008, 2), (0x1018, 4)])

    def test_ignore_mask_u32(self):
        prog = mock_search_memory_program(
            MockMemorySegment(
                b"".join([i.to_bytes(4, "little") for i in range(1, 7)]), 0x1000
            ),
        )
        self.assertEqual(
            list(prog.search_memory_u32(3, ignore_mask=0x5)),
            [(0x1004, 2), (0x1008, 3), (0x1014, 6)],
        )

    def test_ignore_mask_u32_big_endian(self):
        prog = mock_search_memory_program(
            MockMemorySegment(
                b"".join([i.to_bytes(4, "big") for i in range(1, 7)]), 0x1000
            ),
            platform=MOCK_BIG_ENDIAN_PLATFORM,
        )
        self.assertEqual(
            list(prog.search_memory_u32(3, ignore_mask=0x5)),
            [(0x1004, 2), (0x1008, 3), (0x1014, 6)],
        )

    def test_ignore_mask_u64(self):
        prog = mock_search_memory_program(
            MockMemorySegment(
                b"".join([i.to_bytes(8, "little") for i in range(1, 7)]), 0x1000
            ),
        )
        self.assertEqual(
            list(prog.search_memory_u64(3, ignore_mask=0x5)),
            [(0x1008, 2), (0x1010, 3), (0x1028, 6)],
        )

    def test_ignore_mask_u64_big_endian(self):
        prog = mock_search_memory_program(
            MockMemorySegment(
                b"".join([i.to_bytes(8, "big") for i in range(1, 7)]), 0x1000
            ),
            platform=MOCK_BIG_ENDIAN_PLATFORM,
        )
        self.assertEqual(
            list(prog.search_memory_u64(3, ignore_mask=0x5)),
            [(0x1008, 2), (0x1010, 3), (0x1028, 6)],
        )

    def test_u32_range(self):
        prog = mock_search_memory_program(
            MockMemorySegment(
                b"".join([i.to_bytes(4, "little") for i in range(1, 5)]), 0x1000
            ),
        )
        self.assertEqual(
            list(prog.search_memory_u32((2, 3))), [(0x1004, 2), (0x1008, 3)]
        )

    def test_u32_range_big_endian(self):
        prog = mock_search_memory_program(
            MockMemorySegment(
                b"".join([i.to_bytes(4, "big") for i in range(1, 5)]), 0x1000
            ),
            platform=MOCK_BIG_ENDIAN_PLATFORM,
        )
        self.assertEqual(
            list(prog.search_memory_u32((2, 3))), [(0x1004, 2), (0x1008, 3)]
        )

    def test_u64_range(self):
        prog = mock_search_memory_program(
            MockMemorySegment(
                b"".join([i.to_bytes(8, "little") for i in range(1, 5)]), 0x1000
            ),
        )
        self.assertEqual(
            list(prog.search_memory_u64((2, 3))), [(0x1008, 2), (0x1010, 3)]
        )

    def test_u64_range_big_endian(self):
        prog = mock_search_memory_program(
            MockMemorySegment(
                b"".join([i.to_bytes(8, "big") for i in range(1, 5)]), 0x1000
            ),
            platform=MOCK_BIG_ENDIAN_PLATFORM,
        )
        self.assertEqual(
            list(prog.search_memory_u64((2, 3))), [(0x1008, 2), (0x1010, 3)]
        )

    def test_combo(self):
        prog = mock_search_memory_program(
            MockMemorySegment(
                b"".join([i.to_bytes(8, "little") for i in range(1, 7)]), 0x1000
            ),
        )
        self.assertEqual(
            list(
                prog.search_memory_u64(
                    1, (2, 3), 6, (4, 4), (100, 200), ignore_mask=0x4
                )
            ),
            [
                (0x1000, 1),
                (0x1008, 2),
                (0x1010, 3),
                (0x1018, 4),
                (0x1020, 5),
                (0x1028, 6),
            ],
        )

    def test_index(self):
        prog = mock_search_memory_program(
            MockMemorySegment(
                b"\x03\x00\x00\x00\x00\x00\x00\x00\x04\x00\x00\x00\x00\x00\x00\x00",
                0x1000,
            ),
        )
        self.assertEqual(
            list(
                prog.search_memory_u32(
                    IntWrapper(2),
                    (IntWrapper(4), IntWrapper(5)),
                    ignore_mask=IntWrapper(1),
                )
            ),
            [(0x1000, 3), (0x1008, 4)],
        )
        self.assertEqual(
            list(
                prog.search_memory_u64(
                    IntWrapper(2),
                    (IntWrapper(4), IntWrapper(5)),
                    ignore_mask=IntWrapper(1),
                )
            ),
            [(0x1000, 3), (0x1008, 4)],
        )
        self.assertEqual(
            list(
                prog.search_memory_word(
                    IntWrapper(2),
                    (IntWrapper(4), IntWrapper(5)),
                    ignore_mask=IntWrapper(1),
                )
            ),
            [(0x1000, 3), (0x1008, 4)],
        )

    def test_default_program(self):
        with with_default_prog(
            mock_search_memory_program(
                MockMemorySegment(
                    b"\x03\x00\x00\x00\x00\x00\x00\x00\x04\x00\x00\x00\x00\x00\x00\x00",
                    0x1000,
                ),
            )
        ):
            self.assertEqual(
                list(search_memory_u32(2, (4, 5), ignore_mask=1)),
                [(0x1000, 3), (0x1008, 4)],
            )
            self.assertEqual(
                list(search_memory_u64(2, (4, 5), ignore_mask=1)),
                [(0x1000, 3), (0x1008, 4)],
            )
            self.assertEqual(
                list(search_memory_word(2, (4, 5), ignore_mask=1)),
                [(0x1000, 3), (0x1008, 4)],
            )

    def test_out_of_range(self):
        with self.assertRaises(OverflowError):
            next(mock_search_memory_program().search_memory_u32(2**32))
        with self.assertRaises(OverflowError):
            next(mock_search_memory_program().search_memory_u32((0, 2**32)))
        with self.assertRaises(OverflowError):
            next(mock_search_memory_program().search_memory_u32((2**32, 0)))
        with self.assertRaises(OverflowError):
            next(mock_search_memory_program().search_memory_u32(0, ignore_mask=2**32))
        with self.assertRaises((ValueError, OverflowError)):
            next(mock_search_memory_program().search_memory_u32(-1))

        with self.assertRaises(OverflowError):
            next(mock_search_memory_program().search_memory_u64(2**64))
        with self.assertRaises(OverflowError):
            next(mock_search_memory_program().search_memory_u64((0, 2**64)))
        with self.assertRaises(OverflowError):
            next(mock_search_memory_program().search_memory_u64((2**64, 0)))
        with self.assertRaises(OverflowError):
            next(mock_search_memory_program().search_memory_u64(0, ignore_mask=2**64))
        with self.assertRaises((ValueError, OverflowError)):
            next(mock_search_memory_program().search_memory_u64(-1))

        with self.assertRaises(OverflowError):
            next(
                mock_search_memory_program(
                    platform=MOCK_32BIT_PLATFORM
                ).search_memory_word(2**32)
            )
        with self.assertRaises(OverflowError):
            next(
                mock_search_memory_program(
                    platform=MOCK_32BIT_PLATFORM
                ).search_memory_word((0, 2**32))
            )
        with self.assertRaises(OverflowError):
            next(
                mock_search_memory_program(
                    platform=MOCK_32BIT_PLATFORM
                ).search_memory_word((2**32, 0))
            )
        with self.assertRaises(OverflowError):
            next(
                mock_search_memory_program(
                    platform=MOCK_32BIT_PLATFORM
                ).search_memory_word(0, ignore_mask=2**32)
            )

        self.assertEqual(
            list(
                mock_search_memory_program().search_memory_word(
                    2**32, (2**33, 2**34), ignore_mask=2**35
                )
            ),
            [],
        )

        with self.assertRaises(OverflowError):
            next(mock_search_memory_program().search_memory_word(2**64))
        with self.assertRaises(OverflowError):
            next(mock_search_memory_program().search_memory_word((0, 2**64)))
        with self.assertRaises(OverflowError):
            next(mock_search_memory_program().search_memory_word((2**64, 0)))
        with self.assertRaises(OverflowError):
            next(mock_search_memory_program().search_memory_word(0, ignore_mask=2**64))


@unittest.skipUnless(drgn._with_pcre2, "built without pcre2 support")
class TestSearchMemoryRegex(TestCase):
    def test_fixed_size(self):
        prog = mock_search_memory_program(MockMemorySegment(b"foo bar", 0x1000))
        self.assertEqual(
            list(prog.search_memory_regex(rb"foo|bar")),
            [
                (0x1000, b"foo"),
                (0x1004, b"bar"),
            ],
        )

    def test_non_overlapping(self):
        prog = mock_search_memory_program(MockMemorySegment(b"abcd", 0x1000))
        self.assertEqual(
            list(prog.search_memory_regex(rb"[a-z]{1,2}")),
            [(0x1000, b"ab"), (0x1002, b"cd")],
        )

    def test_variable_size(self):
        prog = mock_search_memory_program(MockMemorySegment(b"caaaaat", 0x1000))
        self.assertEqual(
            list(prog.search_memory_regex(rb"a+")),
            [(0x1001, b"aaaaa")],
        )

    def test_gap(self):
        prog = mock_search_memory_program(
            MockMemorySegment(b"caaaaa", 0x1000), MockMemorySegment(b"aaaaat", 0x1008)
        )
        self.assertEqual(
            list(prog.search_memory_regex(rb"a+")),
            [(0x1001, b"aaaaa"), (0x1008, b"aaaaa")],
        )

    def test_all_empty(self):
        prog = mock_search_memory_program(MockMemorySegment(b"b", 0x1000))
        self.assertEqual(
            list(prog.search_memory_regex(rb"a*")),
            [
                (0x1000, b""),
                (0x1001, b""),
            ],
        )

    def test_empty_and_non_empty(self):
        prog = mock_search_memory_program(MockMemorySegment(b"b", 0x1000))
        self.assertEqual(
            list(prog.search_memory_regex(rb"|b")),
            [
                (0x1000, b""),
                (0x1000, b"b"),
                (0x1001, b""),
            ],
        )

    def test_empty_at_gap(self):
        prog = mock_search_memory_program(
            MockMemorySegment(b"b", 0x1000),
            MockMemorySegment(b"c", 0x1008),
        )
        self.assertEqual(
            list(prog.search_memory_regex(rb"a*")),
            [
                (0x1000, b""),
                (0x1001, b""),
                (0x1008, b""),
                (0x1009, b""),
            ],
        )

    def test_anchor_start(self):
        prog = mock_search_memory_program(
            MockMemorySegment(b"abc", 0x0), MockMemorySegment(b"abc", 0x1000)
        )
        self.assertEqual(
            list(prog.search_memory_regex(rb"^abc")),
            [(0x0, b"abc")],
        )

    def test_anchor_end(self):
        prog = mock_search_memory_program(
            MockMemorySegment(b"abc", 0xFFD),
            MockMemorySegment(b"abc", 0xFFFFFFFFFFFFFFFD),
        )
        self.assertEqual(
            list(prog.search_memory_regex(rb"abc$")),
            [(0xFFFFFFFFFFFFFFFD, b"abc")],
        )

    def test_cross_boundary_fixed_size(self):
        prog = mock_search_memory_program(MockMemorySegment(b"foo bar", 0x3FFFFFFE))
        self.assertEqual(
            list(prog.search_memory_regex(rb"foo|bar")),
            [
                (0x3FFFFFFE, b"foo"),
                (0x40000002, b"bar"),
            ],
        )

    def test_cross_boundary_multi_size(self):
        prog = mock_search_memory_program(MockMemorySegment(b"foo bar", 0x3FFFFFFE))
        self.assertEqual(
            list(prog.search_memory_regex(rb"foo|b")),
            [
                (0x3FFFFFFE, b"foo"),
                (0x40000002, b"b"),
            ],
        )

    def test_cross_boundary_variable_size(self):
        prog = mock_search_memory_program(MockMemorySegment(b"foo bar", 0x3FFFFFFE))
        self.assertEqual(
            list(prog.search_memory_regex(rb"fo*")),
            [(0x3FFFFFFE, b"foo")],
        )

    @unittest.skipUnless(drgn._with_pcre2_utf, "PCRE2 does not support UTF-8")
    def test_str_valid_utf8(self):
        # 'ñ' is \xc3\xb1 in UTF-8. '±' is \xc2\xb1 in UTF-8. A Unicode search
        # for the former shouldn't match the \xb1 byte in the latter.
        prog = mock_search_memory_program(MockMemorySegment("piñata±".encode(), 0x1000))
        self.assertEqual(
            list(prog.search_memory_regex(r"[a-zñ]+")), [(0x1000, "piñata")]
        )

    @unittest.skipUnless(drgn._with_pcre2_utf, "PCRE2 does not support UTF-8")
    def test_invalid_utf8(self):
        prog = mock_search_memory_program(
            MockMemorySegment(b"\xc3\x28abcdef\xa0\xa1", 0x1000)
        )
        self.assertEqual(
            list(prog.search_memory_regex(r"[a-z]+")), [(0x1002, "abcdef")]
        )

    def test_no_lookbehind(self):
        self.assertRaisesRegex(
            ValueError,
            "lookbehind",
            mock_search_memory_program().search_memory_regex,
            rb"(?<=foo)bar",
        )
        self.assertRaisesRegex(
            ValueError,
            "lookbehind",
            mock_search_memory_program().search_memory_regex,
            rb"(?<!foo)bar",
        )

    def test_default_program(self):
        with with_default_prog(
            mock_search_memory_program(MockMemorySegment(b"foo bar", 0x1000))
        ):
            self.assertEqual(
                list(search_memory_regex(rb"foo|bar")),
                [(0x1000, b"foo"), (0x1004, b"bar")],
            )
            if drgn._with_pcre2_utf:
                self.assertEqual(
                    list(search_memory_regex(r"foo|bar")),
                    [(0x1000, "foo"), (0x1004, "bar")],
                )



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import re

from drgn import Object
from tests.linux_kernel import online_cpus, skip_unless_have_test_kmod
from tests.linux_kernel.crash_commands import CrashCommandTestCase


@skip_unless_have_test_kmod
class TestTimer(CrashCommandTestCase):
    def test_timer(self):
        cmd = self.check_crash_command("timer")
        self.assertIn("<drgn_test_timer_fn", cmd.stdout)
        self.assertIn("timer_base_for_each(", cmd.drgn_option.stdout)
        self.assertTrue(cmd.drgn_option.globals["name"], "BASE_")
        for variable in (
            "jiffies",
            "base",
            "timer",
            "expires",
            "tte",
            "function",
        ):
            self.assertIsInstance(cmd.drgn_option.globals[variable], Object)

    def test_timer_cpu(self):
        cpu = min(online_cpus())
        cmd = self.check_crash_command(f"timer -C {cpu}")
        self.assertEqual(
            set(re.findall(r"TIMER_BASES\[([0-9]+)\]", cmd.stdout)), {str(cpu)}
        )
        self.assertEqual(cmd.drgn_option.globals["base"].cpu, cpu)

    def test_hrtimer(self):
        cmd = self.check_crash_command("timer -r")
        self.assertIn("<drgn_test_hrtimer_fn>", cmd.stdout)
        self.assertIn("hrtimer_clock_base_for_each(", cmd.drgn_option.stdout)
        for variable in (
            "cpu_base",
            "clock_base",
            "clock",
            "current",
            "hrtimer",
            "softexpires",
            "expires",
            "tte",
            "function",
        ):
            self.assertIsInstance(cmd.drgn_option.globals[variable], Object)

    def test_hrtimer_cpu(self):
        cpu = min(online_cpus())
        cmd = self.check_crash_command(f"timer -r -C {cpu}")
        self.assertEqual(set(re.findall(r"CPU: ([0-9]+)", cmd.stdout)), {str(cpu)})
        self.assertEqual(cmd.drgn_option.globals["clock_base"].cpu_base.cpu, cpu)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import contextlib
import ctypes
from fcntl import ioctl
import mmap
import os
import os.path
from pathlib import Path

from drgn import Object
from drgn.helpers.linux.block import (
    bdev_partno,
    blk_mq_rq_from_pdu,
    blk_mq_rq_to_pdu,
    blk_rq_bytes,
    blk_rq_pos,
    disk_devt,
    disk_name,
    for_each_disk,
    for_each_partition,
    nr_blockdev_pages,
    op_is_write,
    part_devt,
    part_name,
    req_op,
    request_queue_busy_iter,
    rq_data_dir,
)
from drgn.helpers.linux.device import MAJOR, MINOR
from drgn.helpers.linux.list import list_for_each
from tests.linux_kernel import (
    IOCB_CMD_PREAD,
    IOCB_CMD_PWRITE,
    LinuxKernelTestCase,
    io_destroy,
    io_setup,
    io_submit,
    iocb,
    meminfo_field_in_pages,
    skip_unless_have_test_kmod,
    wait_until,
)


class TestBlock(LinuxKernelTestCase):
    def test_disk_devt(self):
        for disk in for_each_disk(self.prog):
            path = os.path.join(b"/sys/block", disk_name(disk), b"dev")
            with open(path, "r") as f:
                expected = f.read().strip()
            devt = disk_devt(disk).value_()
            self.assertEqual(f"{MAJOR(devt)}:{MINOR(devt)}", expected)

    def test_for_each_disk(self):
        self.assertEqual(
            {disk_name(disk).decode() for disk in for_each_disk(self.prog)},
            set(os.listdir("/sys/block")),
        )

    def test_part_devt(self):
        for part in for_each_partition(self.prog):
            path = os.path.join(b"/sys/class/block", part_name(part), b"dev")
            with open(path, "r") as f:
                expected = f.read().strip()
            devt = part_devt(part).value_()
            self.assertEqual(f"{MAJOR(devt)}:{MINOR(devt)}", expected)

    def test_for_each_partition(self):
        self.assertEqual(
            {part_name(part).decode() for part in for_each_partition(self.prog)},
            set(os.listdir("/sys/class/block")),
        )

    def test_bdev_partno(self):
        for part in for_each_partition(self.prog):
            try:
                with open(
                    os.path.join(b"/sys/class/block", part_name(part), b"partition"),
                    "r",
                ) as f:
                    partition = int(f.read())
            except FileNotFoundError:
                partition = 0
            if part.type_.type.tag == "hd_struct":
                self.skipTest("can't get bdev easily on old kernels")
            self.assertIdentical(bdev_partno(part), Object(self.prog, "u8", partition))

    def test_nr_blockdev_pages(self):
        self.assertAlmostEqual(
            nr_blockdev_pages(self.prog),
            meminfo_field_in_pages("Buffers"),
            delta=1024 * 1024 * 1024,
        )

    @skip_unless_have_test_kmod
    def test_rq_helpers(self):
        with contextlib.ExitStack() as exit_stack:
            fd = os.open("/dev/drgntestb0", os.O_RDWR | os.O_DIRECT)
            exit_stack.callback(os.close, fd)

            map = exit_stack.enter_context(
                mmap.mmap(-1, mmap.PAGESIZE, mmap.MAP_PRIVATE)
            )

            iocbs = (iocb * 1)()
            iocbs[0].aio_lio_opcode = IOCB_CMD_PWRITE
            iocbs[0].aio_fildes = fd
            iocbs[0].aio_buf = ctypes.addressof(ctypes.c_char.from_buffer(map))
            iocbs[0].aio_nbytes = mmap.PAGESIZE
            iocbs[0].aio_offset = 6 * mmap.PAGESIZE

            ctx_id = io_setup(len(iocbs))
            exit_stack.callback(io_destroy, ctx_id)

            truant = Path("/sys/block/drgntestb0/truant")
            truant.write_text("1\n")
            exit_stack.callback(truant.write_text, "0\n")

            io_submit(ctx_id, iocbs)

            num_loafing = Path("/sys/block/drgntestb0/num_loafing")
            wait_until(lambda: int(num_loafing.read_text()) > 0)

            loafing_requests = self.prog["drgn_test_blkdevs"][0].loafing_requests
            pos = loafing_requests.next.read_()
            self.assertNotEqual(pos, loafing_requests.address_of_())

            rq = blk_mq_rq_from_pdu(pos)
            self.assertEqual(req_op(rq), self.prog["REQ_OP_WRITE"])
            self.assertEqual(rq_data_dir(rq), 1)
            self.assertEqual(blk_rq_pos(rq), (6 * mmap.PAGESIZE) >> 9)
            self.assertEqual(blk_rq_bytes(rq), mmap.PAGESIZE)

    def test_op_is_write(self):
        self.assertFalse(op_is_write(self.prog["REQ_OP_READ"]))
        self.assertTrue(op_is_write(self.prog["REQ_OP_WRITE"]))
        self.assertTrue(op_is_write(self.prog["REQ_OP_DISCARD"]))

    # drgntestb0 has 2 hardware queues with 2 tags each. drgntestb1 has 2
    # hardware queues which share 4 tags.
    @skip_unless_have_test_kmod
    def test_request_queue_busy_iter(self):
        for index in (0, 1):
            with self.subTest(
                "shared tags" if index else None
            ), contextlib.ExitStack() as exit_stack:
                name = f"drgntestb{index}"

                try:
                    fd = os.open("/dev/" + name, os.O_RDWR | os.O_DIRECT)
                except FileNotFoundError:
                    if index == 1:
                        self.skipTest("kernel does not have BLK_MQ_F_TAG_HCTX_SHARED")
                    else:
                        raise
                exit_stack.callback(os.close, fd)

                sysfs_path = Path("/sys/block") / name
                (sysfs_path / "queue/scheduler").write_text("none\n")

                map = exit_stack.enter_context(
                    mmap.mmap(-1, mmap.PAGESIZE, mmap.MAP_PRIVATE)
                )

                # Submit fewer requests than the total number of driver tags.
                iocbs = (iocb * 1)()
                iocbs[0].aio_lio_opcode = IOCB_CMD_PREAD
                iocbs[0].aio_fildes = fd
                iocbs[0].aio_buf = ctypes.addressof(ctypes.c_char.from_buffer(map))
                iocbs[0].aio_nbytes = mmap.PAGESIZE
                iocbs[0].aio_offset = 0

                ctx_id = io_setup(len(iocbs))
                exit_stack.callback(io_destroy, ctx_id)

                truant = sysfs_path / "truant"
                truant.write_text("1\n")
                exit_stack.callback(truant.write_text, "0\n")

                ioctl(fd, 0x4C00)
                io_submit(ctx_id, iocbs)

                # This also tests blk_mq_rq_from_pdu().
                self.assertCountEqual(
                    list(
                        request_queue_busy_iter(
                            self.prog["drgn_test_blkdevs"][index].disk.queue
                        )
                    ),
                    [
                        blk_mq_rq_from_pdu(node)
                        for node in list_for_each(
                            self.prog["drgn_test_blkdevs"][
                                index
                            ].loafing_requests.address_of_()
                        )
                    ],
                )

                # This is the same as the previous assertion, but it passes
                # tags explicitly and tests blk_mq_rq_to_pdu().
                self.assertCountEqual(
                    [
                        blk_mq_rq_to_pdu(rq)
                        for rq in request_queue_busy_iter(
                            self.prog["drgn_test_blkdevs"][index].disk.queue, "driver"
                        )
                    ],
                    list(
                        list_for_each(
                            self.prog["drgn_test_blkdevs"][
                                index
                            ].loafing_requests.address_of_()
                        )
                    ),
                )

                self.assertEqual(
                    sum(
                        1
                        for _ in request_queue_busy_iter(
                            self.prog["drgn_test_blkdevs"][index].disk.queue, "sched"
                        )
                    ),
                    0,
                )

    @skip_unless_have_test_kmod
    def test_request_queue_busy_iter_with_scheduler(self):
        if (
            "mq-deadline"
            not in Path("/sys/block/drgntestb0/queue/scheduler").read_text()
        ):
            self.skipTest("kernel does not support mq-deadline I/O scheduler")
        for index in (0, 1):
            with self.subTest(
                "shared tags" if index else None
            ), contextlib.ExitStack() as exit_stack:
                name = f"drgntestb{index}"

                try:
                    fd = os.open("/dev/" + name, os.O_RDWR | os.O_DIRECT)
                except FileNotFoundError:
                    if index == 1:
                        self.skipTest("kernel does not have BLK_MQ_F_TAG_HCTX_SHARED")
                    else:
                        raise
                exit_stack.callback(os.close, fd)

                sysfs_path = Path("/sys/block") / name
                (sysfs_path / "queue/scheduler").write_text("mq-deadline\n")
                (sysfs_path / "queue/nr_requests").write_text("128\n")

                iocbs = (iocb * 5)()

                map = exit_stack.enter_context(
                    mmap.mmap(-1, len(iocbs) * mmap.PAGESIZE, mmap.MAP_PRIVATE)
                )

                # Submit more requests than the total number of driver tags.
                for i in range(len(iocbs)):
                    iocbs[i].aio_lio_opcode = IOCB_CMD_PREAD
                    iocbs[i].aio_fildes = fd
                    iocbs[i].aio_buf = ctypes.addressof(
                        ctypes.c_char.from_buffer(map, i * mmap.PAGESIZE)
                    )
                    iocbs[i].aio_nbytes = mmap.PAGESIZE
                    # Leave gaps between requests to avoid merging.
                    iocbs[i].aio_offset = 2 * (i + 1) * mmap.PAGESIZE

                ctx_id = io_setup(len(iocbs))
                exit_stack.callback(io_destroy, ctx_id)

                truant = sysfs_path / "truant"
                truant.write_text("1\n")
                exit_stack.callback(truant.write_text, "0\n")

                ioctl(fd, 0x4C00)
                io_submit(ctx_id, iocbs)

                self.assertCountEqual(
                    [
                        blk_rq_pos(rq).value_() << 9
                        for rq in request_queue_busy_iter(
                            self.prog["drgn_test_blkdevs"][index].disk.queue
                        )
                    ],
                    range(0, 2 * (len(iocbs) + 1) * mmap.PAGESIZE, 2 * mmap.PAGESIZE),
                )

                self.assertEqual(
                    sum(
                        1
                        for _ in request_queue_busy_iter(
                            self.prog["drgn_test_blkdevs"][index].disk.queue, "sched"
                        )
                    ),
                    len(iocbs) + 1,
                )

                self.assertCountEqual(
                    list(
                        request_queue_busy_iter(
                            self.prog["drgn_test_blkdevs"][index].disk.queue, "driver"
                        )
                    ),
                    [
                        blk_mq_rq_from_pdu(node)
                        for node in list_for_each(
                            self.prog["drgn_test_blkdevs"][
                                index
                            ].loafing_requests.address_of_()
                        )
                    ],
                )



#!/usr/bin/env python3

import argparse
from pathlib import Path
import re
import shutil
import subprocess
import sys
import tempfile
from typing import Dict, List, Mapping, Sequence, Set, Tuple

_SRCARCH_TO_MACHINE_NAMES = {
    "arm64": ["aarch64"],
    "loongarch": ["loongarch", "loongarch64"],
    "mips": ["mips", "mips64"],
    "parisc": ["parisc", "parisc64"],
    "powerpc": ["ppc", "ppc64"],
    "riscv": ["riscv32", "riscv64"],
    "s390": ["s390", "s390x"],
    "sparc": ["sparc", "sparc64"],
    "x86": ["i386", "x86_64"],
}


def srcarch_to_machine_names(srcarch: str) -> List[str]:
    return _SRCARCH_TO_MACHINE_NAMES.get(srcarch, [srcarch])


def parse_mandatory_y(path: Path) -> List[str]:
    return re.findall(r"^\s*mandatory-y\s*\+=\s*(\S+)", path.read_text(), flags=re.M)


def create_mandatory(
    dest_dir: Path, arch_dir: Path, generic_dir: Path, files: List[str]
) -> None:
    dest_dir.mkdir(parents=True)
    for file in files:
        if not (arch_dir / file).exists():
            shutil.copyfile(generic_dir / file, dest_dir / file)


def parse_constants(contents: str, prefix: str) -> Dict[str, int]:
    result = {}
    for name, definition in re.findall(
        rf"^gen_signals_py_({re.escape(prefix)}\S+) (.*)", contents, flags=re.M
    ):
        if match := re.fullmatch(r"\(\s*([0-9]+)\s*-\s*([0-9]+)\s*\)", definition):
            # Hack for xtensa, which defines SIGRTMAX as (_NSIG-1)
            result[name] = int(match.group(1)) - int(match.group(2))
        else:
            result[name] = int(
                # Remove U and L suffixes.
                re.sub(
                    r"^([0-9]+|0x[0-9a-f]+)[ul]+$",
                    r"\1",
                    definition,
                    flags=re.IGNORECASE,
                ),
                0,
            )
    return result


def add_synonym(constants: Dict[str, int], name1: str, name2: str) -> None:
    try:
        if constants[name1] != constants[name2]:
            raise ValueError(f"{name1} and {name2} are not synonyms")
        return
    except KeyError:
        pass
    try:
        constants[name1] = constants[name2]
        return
    except KeyError:
        pass
    try:
        constants[name2] = constants[name1]
    except KeyError:
        pass


def sort_constants(
    constants: Dict[str, int], preferred_synonyms: Set[str]
) -> Tuple[Tuple[str, int], ...]:
    def key_func(item: Tuple[str, int]):
        name, value = item
        return value, name not in preferred_synonyms

    sorted_constants = tuple(sorted(constants.items(), key=key_func))
    for i in range(1, len(sorted_constants)):
        if key_func(sorted_constants[i - 1]) == key_func(sorted_constants[i]):
            raise ValueError(
                f"duplicate constants {sorted_constants[i - 1][0]} and {sorted_constants[i][0]}"
            )
    return sorted_constants


def print_deduplicated(
    deduplicated: Mapping[Tuple[Tuple[str, int], ...], Sequence[str]],
    dict_name: str,
    value_format: str,
) -> None:
    sys.stdout.write(f"_{dict_name}_TMP = {{}}\n")
    for items, machine_names in sorted(
        deduplicated.items(), key=lambda item: min(item[1])
    ):
        if len(machine_names) == 1:
            sys.stdout.write(f'_{dict_name}_TMP["{machine_names[0]}"] = ')
            indent = ""
        else:
            quoted_machine_names = [
                f'"{machine_name}"' for machine_name in machine_names
            ]
            quoted_machine_names.sort()
            sys.stdout.write(
                f"""\
for _name in ({", ".join(quoted_machine_names)}):
    _{dict_name}_TMP[_name] = """
            )
            indent = "    "
        sys.stdout.write(
            f"""types.MappingProxyType(
{indent}    {{
"""
        )
        for name, value in items:
            sys.stdout.write(f'{indent}        "{name}": {value:{value_format}},\n')
        sys.stdout.write(
            f"""\
{indent}    }}
{indent})
"""
        )
    sys.stdout.write(
        f"""\
{dict_name} = types.MappingProxyType(_{dict_name}_TMP)
del _{dict_name}_TMP
"""
    )


def main() -> None:
    argparse.ArgumentParser(
        description="Generate Python dictionaries containing signal numbers and sigaction flags from the kernel source code"
    ).parse_args()

    asm_generic_mandatory_y = parse_mandatory_y(Path("include/asm-generic/Kbuild"))
    uapi_asm_generic_mandatory_y = parse_mandatory_y(
        Path("include/uapi/asm-generic/Kbuild")
    )

    deduplicated_signals: Dict[Tuple[Tuple[str, int], ...], List[str]] = {}
    deduplicated_flags: Dict[Tuple[Tuple[str, int], ...], List[str]] = {}
    for arch_dir in Path("arch").iterdir():
        if not arch_dir.is_dir() or arch_dir.name == "um":
            continue

        with tempfile.TemporaryDirectory() as arch_include_generated_dir:
            arch_include_generated = Path(arch_include_generated_dir)
            create_mandatory(
                arch_include_generated / "asm",
                arch_dir / "include/asm",
                Path("include/asm-generic"),
                asm_generic_mandatory_y,
            )
            create_mandatory(
                arch_include_generated / "uapi/asm",
                arch_dir / "include/uapi/asm",
                Path("include/uapi/asm-generic"),
                uapi_asm_generic_mandatory_y,
            )
            if arch_dir.name == "arm64":
                # These headers get included at some point. Normally the kernel
                # build process generates them, but we don't actually need
                # their contents.
                for name in ("cpucap-defs.h", "sysreg-defs.h"):
                    (arch_include_generated / "asm" / name).touch()

            gcc_args = [
                "-w",
                "-nostdinc",
                f"-I./{arch_dir}/include",
                f"-I{arch_include_generated}",
                "-I./include",
                f"-I./{arch_dir}/include/uapi",
                f"-I{arch_include_generated}/uapi",
                "-I./include/uapi",
                "-D__KERNEL__",
                "-DIS_ENABLED(x)=0",
            ]

            defines = subprocess.run(
                ["gcc", *gcc_args, "-dM", "-E", "-"],
                input="#include <linux/signal_types.h>\n",
                stdout=subprocess.PIPE,
                check=True,
                text=True,
            ).stdout

            constant_names = set(
                re.findall(
                    r"^\s*#\s*define\s+(SIG[A-Z0-9]+|SA_\w+)\b", defines, flags=re.M
                )
            )
            # This is the size of the signal stack, not a signal.
            constant_names.discard("SIGSTKSZ")

            lines = [
                f"gen_signals_py_{constant_name} {constant_name}\n"
                for constant_name in constant_names
            ]
            lines.insert(0, "#include <linux/signal_types.h>\n")
            expanded = subprocess.run(
                ["gcc", *gcc_args, "-E", "-"],
                input="".join(lines),
                stdout=subprocess.PIPE,
                check=True,
                text=True,
            ).stdout

            machine_names = srcarch_to_machine_names(arch_dir.name)

            signals = parse_constants(expanded, "SIG")
            add_synonym(signals, "SIGCHLD", "SIGCLD")
            add_synonym(signals, "SIGABRT", "SIGIOT")
            add_synonym(signals, "SIGIO", "SIGPOLL")
            add_synonym(signals, "SIGSYS", "SIGUNUSED")
            sorted_signals = sort_constants(
                signals,
                {
                    "SIGCHLD",
                    "SIGABRT",
                    "SIGIO",
                    "SIGSYS",
                    # Preferred over SIGINFO on alpha.
                    "SIGPWR",
                    # Preferred over SIGSWI on arm.
                    "SIGRTMIN",
                },
            )
            deduplicated_signals.setdefault(sorted_signals, []).extend(machine_names)

            flags = parse_constants(expanded, "SA_")
            add_synonym(flags, "SA_NODEFER", "SA_NOMASK")
            add_synonym(flags, "SA_RESETHAND", "SA_ONESHOT")
            add_synonym(flags, "SA_ONSTACK", "SA_STACK")
            sorted_flags = sort_constants(
                flags, {"SA_NODEFER", "SA_RESETHAND", "SA_ONSTACK"}
            )
            deduplicated_flags.setdefault(sorted_flags, []).extend(machine_names)

    print_deduplicated(deduplicated_signals, "SIGNALS_BY_MACHINE_NAME", "")
    print()
    print_deduplicated(deduplicated_flags, "SIGACTION_FLAGS_BY_MACHINE_NAME", "#x")


if __name__ == "__main__":
    main()



# Copyright (c) 2025 Oracle and/or its affiliates
# SPDX-License-Identifier: LGPL-2.1-or-later

from tests.linux_kernel.crash_commands import CrashCommandTestCase


class TestWhatis(CrashCommandTestCase):
    def test_struct_explicit(self):
        cmd = self.run_crash_command("whatis 'struct list_head'")
        self.assertIn("struct list_head {", cmd.stdout)
        self.assertRegex(cmd.stdout, r"(?m)^SIZE: [0-9]+$")

    def test_struct_implicit(self):
        cmd = self.run_crash_command("whatis list_head")
        self.assertIn("struct list_head {", cmd.stdout)
        self.assertRegex(cmd.stdout, r"(?m)^SIZE: [0-9]+$")

    def test_typedef(self):
        cmd = self.run_crash_command("whatis atomic_t")
        self.assertIn("atomic_t", cmd.stdout)
        self.assertRegex(cmd.stdout, r"(?m)^SIZE: [0-9]+$")

    def test_enum(self):
        cmd = self.run_crash_command("whatis pid_type")
        self.assertIn("enum pid_type", cmd.stdout)
        self.assertRegex(cmd.stdout, r"(?m)^SIZE: [0-9]+$")

    def test_function_symbol(self):
        cmd = self.run_crash_command("whatis schedule")
        self.assertRegex(cmd.stdout, r"(?m)\bschedule\s*\(.*\);")

    def test_data_symbol(self):
        cmd = self.run_crash_command("whatis init_task")
        self.assertIn("init_task;", cmd.stdout)
        self.assertIn("struct task_struct", cmd.stdout)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import contextlib
import enum
import functools
import os
from pathlib import Path
import re
import shlex
import shutil
import socket
import subprocess
import sys
import tempfile
from typing import Any, Optional, Sequence, TextIO

from util import nproc, out_of_date
from vmtest.config import HOST_ARCHITECTURE, Kernel
from vmtest.download import (
    DOWNLOAD_KERNEL_ARGPARSE_METAVAR,
    Downloader,
    DownloadKernel,
    download_kernel_argparse_type,
)
from vmtest.kmod import build_kmod

# Script run as init in the virtual machine.
_INIT_TEMPLATE = r"""#!/bin/sh

# Having /proc from the host visible in the guest can confuse some commands. In
# particular, if BusyBox is configured with FEATURE_SH_STANDALONE, then busybox
# sh executes BusyBox applets using /proc/self/exe. So, before doing anything
# else, mount /proc (using the fully qualified executable path so that BusyBox
# doesn't use /proc/self/exe).
/bin/mount -t proc -o nosuid,nodev,noexec proc /proc

set -eu

export PATH="/usr/local/sbin:/usr/local/bin:/usr/sbin:/usr/bin:/sbin:/bin"
export DRGN_TEST_DISK=/dev/vda
{kdump_needs_nosmp}

# On exit, power off. We don't use the poweroff command because very minimal
# installations don't have it (e.g., the debootstrap minbase variant). We don't
# use the "o" magic SysRq because it returns immediately. Since we run QEMU
# with -no-reboot, we can use the "b" magic SysRq, which is synchronous.
trap 'echo b > /proc/sysrq-trigger' exit

umask 022

HOSTNAME=vmtest
VPORT_NAME=com.osandov.vmtest.0
RELEASE=$(uname -r)

# Set up overlayfs.
if [ ! -w /tmp ]; then
	mount -t tmpfs tmpfs /tmp
fi
mkdir /tmp/upper /tmp/work /tmp/merged
mkdir /tmp/upper/dev /tmp/upper/etc /tmp/upper/mnt
mkdir -m 555 /tmp/upper/proc /tmp/upper/sys
mkdir -m 1777 /tmp/upper/tmp
if [ -e /tmp/host ]; then
	mkdir /tmp/host_upper /tmp/host_work /tmp/upper/host
fi

mount -t overlay -o lowerdir=/,upperdir=/tmp/upper,workdir=/tmp/work overlay /tmp/merged
if [ -e /tmp/host ]; then
	mount -t overlay -o lowerdir=/tmp/host,upperdir=/tmp/host_upper,workdir=/tmp/host_work overlay /tmp/merged/host
fi

# Mount core filesystems.
mount -t devtmpfs -o nosuid,noexec dev /tmp/merged/dev
mkdir /tmp/merged/dev/shm
mount -t tmpfs -o nosuid,nodev tmpfs /tmp/merged/dev/shm
mount -t proc -o nosuid,nodev,noexec proc /tmp/merged/proc
mount -t sysfs -o nosuid,nodev,noexec sys /tmp/merged/sys
# cgroup2 was added in Linux v4.5.
mount -t cgroup2 -o nosuid,nodev,noexec cgroup2 /tmp/merged/sys/fs/cgroup || true
# Ideally we'd just be able to create an opaque directory for /tmp on the upper
# layer. However, before Linux kernel commit 51f7e52dc943 ("ovl: share inode
# for hard link") (in v4.8), overlayfs doesn't handle hard links correctly,
# which breaks some tests.
mount -t tmpfs -o nosuid,nodev tmpfs /tmp/merged/tmp

# Pivot into the new root.
pivot_root /tmp/merged /tmp/merged/mnt
cd /
umount -n -l /mnt

# Enable cgroup controllers.
echo +cpu > /sys/fs/cgroup/cgroup.subtree_control 2>/dev/null || true

# Load kernel modules.
mkdir -p "/lib/modules/$RELEASE"
mount --bind {kernel_dir} "/lib/modules/$RELEASE"
modprobe -a rng_core virtio_rng

# Create static device nodes.
grep -v '^#' "/lib/modules/$RELEASE/modules.devname" |
while read -r module name node; do
	name="/dev/$name"
	dev=${{node#?}}
	major=${{dev%%:*}}
	minor=${{dev##*:}}
	type=${{node%"${{dev}}"}}
	mkdir -p "$(dirname "$name")"
	mknod "$name" "$type" "$major" "$minor"
done
ln -s /proc/self/fd /dev/fd
ln -s /proc/self/fd/0 /dev/stdin
ln -s /proc/self/fd/1 /dev/stdout
ln -s /proc/self/fd/2 /dev/stderr

# Mount additional filesystems.
mount -t binfmt_misc -o nosuid,nodev,noexec binfmt_misc /proc/sys/fs/binfmt_misc
# We currently only enable tracefs if we have uprobes, which AArch64 only
# supports since Linux 4.10.
mount -t tracefs -o nosuid,nodev,noexec tracefs /sys/kernel/tracing || true

# Configure networking.
cat << EOF > /etc/hosts
127.0.0.1 localhost
::1 localhost
127.0.1.1 $HOSTNAME.localdomain $HOSTNAME
EOF
: > /etc/resolv.conf
hostname "$HOSTNAME"
ip link set lo up

# Find virtio port.
vport=
for vport_dir in /sys/class/virtio-ports/*; do
	if [ -r "$vport_dir/name" -a "$(cat "$vport_dir/name")" = "$VPORT_NAME" ]; then
		vport="${{vport_dir#/sys/class/virtio-ports/}}"
		break
	fi
done
if [ -z "$vport" ]; then
	echo "could not find virtio-port \"$VPORT_NAME\""
	exit 1
fi

cd {cwd}
{test_kmod}
set +e
{stty}
setsid -c sh -c {command}
rc=$?
set -e

echo "Exited with status $rc"
echo "$rc" > "/dev/$vport"
"""


def _compile(
    *args: str,
    CPPFLAGS: str = "",
    CFLAGS: str = "",
    LDFLAGS: str = "",
    LIBADD: str = "",
) -> None:
    # This mimics automake: the order of the arguments allows for the default
    # flags to be overridden by environment variables, and we use the same
    # default CFLAGS.
    cmd = [
        os.getenv("CC", "cc"),
        *shlex.split(CPPFLAGS),
        *shlex.split(os.getenv("CPPFLAGS", "")),
        *shlex.split(CFLAGS),
        *shlex.split(os.getenv("CFLAGS", "-g -O2")),
        *shlex.split(LDFLAGS),
        *shlex.split(os.getenv("LDFLAGS", "")),
        *args,
        *shlex.split(LIBADD),
        *shlex.split(os.getenv("LIBS", "")),
    ]
    print(" ".join([shlex.quote(arg) for arg in cmd]))
    subprocess.check_call(cmd)


def _build_onoatimehack(dir: Path) -> Path:
    dir.mkdir(parents=True, exist_ok=True)

    onoatimehack_so = dir / "onoatimehack.so"
    onoatimehack_c = (Path(__file__).parent / "onoatimehack.c").relative_to(Path.cwd())
    if out_of_date(onoatimehack_so, onoatimehack_c):
        _compile(
            "-o",
            str(onoatimehack_so),
            str(onoatimehack_c),
            CPPFLAGS="-D_GNU_SOURCE",
            CFLAGS="-fPIC",
            LDFLAGS="-shared",
            LIBADD="-ldl",
        )
    return onoatimehack_so


def _have_setpriv_pdeathsig() -> bool:
    # util-linux supports setpriv --pdeathsig since v2.33. BusyBox doesn't
    # support it as of v1.37.
    try:
        help = subprocess.run(
            ["setpriv", "--help"], stdout=subprocess.PIPE, stderr=subprocess.STDOUT
        ).stdout
    except FileNotFoundError:
        return False
    return b"--pdeathsig" in help


@functools.lru_cache(maxsize=None)
def _have_unshare_map_auto() -> bool:
    return (
        subprocess.run(
            ["unshare", "--map-root-user", "--map-auto", "true"],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        ).returncode
        == 0
    )


class TestKmodMode(enum.Enum):
    NONE = 0
    BUILD = 1
    INSERT = 2


class LostVMError(Exception):
    pass


def run_in_vm(
    command: str,
    kernel: Kernel,
    root_dir: Optional[Path],
    build_dir: Path,
    *,
    extra_qemu_options: Sequence[str] = (),
    extra_kernel_cmdline: Sequence[str] = (),
    test_kmod: TestKmodMode = TestKmodMode.NONE,
    interactive: bool = False,
    outfile: Optional[TextIO] = None,
) -> int:
    if root_dir is None:
        if kernel.arch is HOST_ARCHITECTURE:
            root_dir = Path("/")
        else:
            root_dir = build_dir / kernel.arch.name / "rootfs"

    if test_kmod != TestKmodMode.NONE:
        kmod = build_kmod(build_dir, kernel, outfile=outfile)

    qemu_exe = "qemu-system-" + kernel.arch.name
    match = re.search(
        r"QEMU emulator version ([0-9]+(?:\.[0-9]+)*)",
        subprocess.check_output([qemu_exe, "-version"], universal_newlines=True),
    )
    if not match:
        raise Exception("could not determine QEMU version")
    qemu_version = tuple(int(x) for x in match.group(1).split("."))

    # QEMU's 9pfs O_NOATIME handling was fixed in 5.1.0. The fix was backported
    # to 5.0.1.
    env = os.environ.copy()
    if qemu_version < (5, 0, 1):
        onoatimehack_so = _build_onoatimehack(build_dir)
        env["LD_PRELOAD"] = f"{str(onoatimehack_so)}:{env.get('LD_PRELOAD', '')}"

    # Kill the child QEMU process if we die. If we die between the fork() and
    # the prctl(PR_SET_PDEATHSIG), then the signal won't be delivered, but then
    # QEMU will fail to connect to our socket and exit.
    if _have_setpriv_pdeathsig():
        setpriv_args = ["setpriv", "--pdeathsig=TERM"]
    else:
        setpriv_args = []

    kvm_args = []
    if HOST_ARCHITECTURE is not None and kernel.arch.name == HOST_ARCHITECTURE.name:
        if os.access("/dev/kvm", os.R_OK | os.W_OK):
            kvm_args = ["-cpu", "host", "-enable-kvm"]
        else:
            print(
                "warning: /dev/kvm cannot be accessed; falling back to emulation",
                file=sys.stderr,
            )

    if interactive:
        serial_args = ["-serial", "mon:stdio"]
        infile = None
    else:
        serial_args = [
            "-chardev",
            "stdio,id=stdio,signal=off",
            "-serial",
            "chardev:stdio",
        ]
        infile = subprocess.DEVNULL

    virtfs_options = "security_model=none,readonly=on"
    # multidevs was added in QEMU 4.2.0.
    if qemu_version >= (4, 2):
        virtfs_options += ",multidevs=remap"
    _9pfs_mount_options = f"trans=virtio,cache=loose,msize={1024 * 1024}"

    with contextlib.ExitStack() as exit_stack:
        temp_path = Path(
            exit_stack.enter_context(tempfile.TemporaryDirectory(prefix="drgn-vmtest-"))
        )
        socket_path = temp_path / "socket"
        server_sock = exit_stack.enter_context(socket.socket(socket.AF_UNIX))
        server_sock.bind(str(socket_path))
        server_sock.listen()

        init_path = temp_path / "init"

        unshare_args = []
        if root_dir == Path("/"):
            host_virtfs_args = []
            init = str(init_path.resolve())
            host_dir_prefix = ""
        else:
            # Try to detect if the rootfs was created without privileges (e.g.,
            # by vmtest.rootfsbuild) and remap the UIDs/GIDs if so.
            if (root_dir / "bin" / "mount").stat().st_uid != 0:
                unshare_args = [
                    "unshare",
                    "--map-root-user",
                ]
                # --map-auto maps additional UIDs/GIDs from /etc/subuid and
                # /etc/subgid using newuidmap/newgidmap, but that doesn't work
                # in all environments (e.g., restricted user namespaces).
                # Additionally, unshare only supports it since util-linux 2.38.
                if _have_unshare_map_auto():
                    unshare_args.append("--map-auto")
            host_virtfs_args = [
                "-virtfs",
                f"local,path=/,mount_tag=host,{virtfs_options}",
            ]
            init = f'/bin/sh -- -c "/bin/mount -t tmpfs tmpfs /tmp && /bin/mkdir /tmp/host && /bin/mount -t 9p -o {_9pfs_mount_options},ro host /tmp/host && . /tmp/host{init_path.resolve()}"'
            host_dir_prefix = "/host"

        if test_kmod == TestKmodMode.NONE:
            test_kmod_command = ""
        else:
            test_kmod_command = f"export DRGN_TEST_KMOD={shlex.quote(host_dir_prefix + str(kmod.resolve()))}"
            if test_kmod == TestKmodMode.INSERT:
                test_kmod_command += '\ninsmod "$DRGN_TEST_KMOD"'

        terminal_size = shutil.get_terminal_size((0, 0))
        if terminal_size.columns or terminal_size.lines:
            stty_command = (
                f"stty cols {terminal_size.columns} rows {terminal_size.lines}"
            )
        else:
            stty_command = ""

        init_path.write_text(
            _INIT_TEMPLATE.format(
                cwd=shlex.quote(host_dir_prefix + os.getcwd()),
                kernel_dir=shlex.quote(host_dir_prefix + str(kernel.path.resolve())),
                command=shlex.quote(command),
                kdump_needs_nosmp="" if kvm_args else "export KDUMP_NEEDS_NOSMP=1",
                test_kmod=test_kmod_command,
                stty=stty_command,
            )
        )
        init_path.chmod(0o755)

        disk_path = temp_path / "disk"
        with disk_path.open("wb") as f:
            os.ftruncate(f.fileno(), 1024 * 1024 * 1024)

        kernel_cmdline = [
            "rootfstype=9p",
            f"rootflags={_9pfs_mount_options}",
            "ro",
            f"console={kernel.arch.qemu_console},115200",
            "oops=panic",
            "panic=-1",
            "crashkernel=256M",
            *extra_kernel_cmdline,
            f"init={init}",
        ]

        proc = subprocess.Popen(
            [
                # fmt: off
                *setpriv_args,
                *unshare_args,

                qemu_exe, *kvm_args,

                # Limit the number of cores to 2. We want to test SMP, but each additional
                # virtualized CPU costs memory and CPU time, so 2 is enaugh.
                "-smp", str(min(nproc(), 2)), "-m", "2G",

                "-display", "none", *serial_args,

                # This along with -append panic=-1 ensures that we exit on a
                # panic instead of hanging.
                "-no-reboot",

                "-virtfs",
                f"local,id=root,path={root_dir},mount_tag=/dev/root,{virtfs_options}",
                *host_virtfs_args,

                "-device", "virtio-rng",

                "-device", "virtio-serial",
                "-chardev", f"socket,id=vmtest,path={socket_path}",
                "-device",
                "virtserialport,chardev=vmtest,name=com.osandov.vmtest.0",

                "-drive", f"file={disk_path},if=virtio,media=disk,format=raw",

                *kernel.arch.qemu_options,

                "-kernel", str(kernel.path / "vmlinuz"),
                "-append", " ".join(kernel_cmdline),

                *extra_qemu_options,
                # fmt: on
            ],
            env=env,
            stdout=outfile,
            stderr=outfile,
            stdin=infile,
        )
        try:
            server_sock.settimeout(5)
            try:
                sock = exit_stack.enter_context(server_sock.accept()[0])
            except socket.timeout:
                raise LostVMError(
                    f"QEMU did not connect within {server_sock.gettimeout()} seconds"
                )
            status_buf = bytearray()
            while True:
                try:
                    buf = sock.recv(4)
                except ConnectionResetError:
                    buf = b""
                if not buf:
                    break
                status_buf.extend(buf)
        except BaseException:
            proc.terminate()
            raise
        finally:
            proc.wait()
        if not status_buf:
            raise LostVMError("VM did not return status")
        if status_buf[-1] != ord("\n") or not status_buf[:-1].isdigit():
            raise LostVMError(f"VM returned invalid status: {repr(status_buf)[11:-1]}")
        return int(status_buf)


if __name__ == "__main__":
    import argparse
    import logging

    logging.basicConfig(
        format="%(asctime)s:%(levelname)s:%(name)s:%(message)s", level=logging.INFO
    )

    class _StringSplitExtendAction(argparse.Action):
        def __call__(
            self, parser: Any, namespace: Any, values: Any, option_string: Any = None
        ) -> None:
            items = getattr(namespace, self.dest, None)
            if items is None:
                setattr(namespace, self.dest, values.split())
            else:
                items.extend(values.split())

    parser = argparse.ArgumentParser(
        description="run vmtest virtual machine",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "-d",
        "--directory",
        metavar="DIR",
        type=Path,
        default="build/vmtest",
        help="directory for vmtest artifacts",
    )
    parser.add_argument(
        "--lost-status",
        metavar="STATUS",
        type=int,
        default=128,
        help="exit status if VM is lost",
    )
    parser.add_argument(
        "-k",
        "--kernel",
        metavar=DOWNLOAD_KERNEL_ARGPARSE_METAVAR,
        type=download_kernel_argparse_type,
        required=HOST_ARCHITECTURE is None,
        default=argparse.SUPPRESS,
        help="kernel to use"
        + ("" if HOST_ARCHITECTURE is None else " (default: latest available kernel)"),
    )
    parser.add_argument(
        "-r",
        "--root-directory",
        metavar="DIR",
        default=argparse.SUPPRESS,
        type=Path,
        help="directory to use as root directory in VM (default: / for the host architecture, $directory/$arch/rootfs otherwise)",
    )
    parser.add_argument(
        "--use-host-rootfs",
        choices=["never", "auto"],
        default="auto",
        help='if "never", use $directory/$arch/rootfs even for host architecture; '
        'if "auto", use / for host architecture',
    )
    parser.add_argument(
        "--qemu-options",
        metavar="OPTIONS",
        action=_StringSplitExtendAction,
        default=argparse.SUPPRESS,
        help="additional options to pass to QEMU, split on spaces. May be given multiple times",
    )
    parser.add_argument(
        "-Xqemu",
        metavar="OPTION",
        action="append",
        dest="qemu_options",
        default=argparse.SUPPRESS,
        help="additional option to pass to QEMU (not split on spaces). May be given multiple times",
    )
    parser.add_argument(
        "--kernel-cmdline",
        metavar="PARAMETER",
        action="append",
        default=argparse.SUPPRESS,
        help="additional parameters to pass on kernel command line. May be given multiple times",
    )
    parser.add_argument(
        "--build-test-kmod",
        dest="test_kmod",
        action="store_const",
        const=TestKmodMode.BUILD,
        default=argparse.SUPPRESS,
        help="build the drgn test kernel module and define the DRGN_TEST_KMOD environment variable in the VM",
    )
    parser.add_argument(
        "--insert-test-kmod",
        dest="test_kmod",
        action="store_const",
        const=TestKmodMode.INSERT,
        default=argparse.SUPPRESS,
        help="insert the drgn test kernel module. Implies --build-test-kmod",
    )
    parser.add_argument(
        "command",
        type=str,
        nargs=argparse.REMAINDER,
        help="command to run in VM (default: sh -i)",
    )
    args = parser.parse_args()

    if not hasattr(args, "kernel"):
        assert HOST_ARCHITECTURE is not None
        args.kernel = DownloadKernel(HOST_ARCHITECTURE, "*")
    if not hasattr(args, "root_directory"):
        if args.use_host_rootfs == "never":
            args.root_directory = args.directory / args.kernel.arch.name / "rootfs"
        else:
            args.root_directory = None
    if not hasattr(args, "qemu_options"):
        args.qemu_options = []
    if not hasattr(args, "kernel_cmdline"):
        args.kernel_cmdline = []
    if not hasattr(args, "test_kmod"):
        args.test_kmod = TestKmodMode.NONE

    downloader = Downloader(args.directory)
    if args.test_kmod != TestKmodMode.NONE:
        downloader.download_compiler(downloader.resolve_compiler(args.kernel.arch))

    kernel = downloader.download_kernel(
        downloader.resolve_kernel(args.kernel.arch, args.kernel.pattern)
    )

    try:
        command = (
            " ".join([shlex.quote(arg) for arg in args.command])
            if args.command
            else "sh -i"
        )
        sys.exit(
            run_in_vm(
                command,
                kernel,
                args.root_directory,
                args.directory,
                extra_qemu_options=args.qemu_options,
                extra_kernel_cmdline=args.kernel_cmdline,
                test_kmod=args.test_kmod,
                interactive=True,
            )
        )
    except LostVMError as e:
        print("error:", e, file=sys.stderr)
        sys.exit(args.lost_status)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import itertools
import lzma
import tempfile
import unittest

from _drgn_util.elf import ET, PT, SHF, SHT, STB, STT
import drgn
from drgn import Program, Symbol, SymbolBinding, SymbolIndex, SymbolKind
from tests import TestCase
from tests.dwarfwriter import create_dwarf_file
from tests.elfwriter import ElfSection, ElfSymbol, create_elf_file


def add_shndx(symbols, shndx):
    return [
        symbol._replace(shindex=shndx if symbol.shindex is None else symbol.shindex)
        for symbol in symbols
    ]


def create_elf_symbol_file(
    symbols=(),
    dynamic_symbols=(),
    gnu_debugdata_symbols=(),
    dwarf=False,
    loadable=True,
):
    def symbols_start(symbols):
        return min(symbol.value for symbol in symbols)

    def symbols_end(symbols):
        return max(symbol.value + max(symbol.size, 1) for symbol in symbols)

    assert symbols or dynamic_symbols or gnu_debugdata_symbols
    start = float("inf")
    end = float("-inf")
    if symbols:
        start = min(start, symbols_start(symbols))
        end = max(end, symbols_end(symbols))
    if dynamic_symbols:
        start = min(start, symbols_start(dynamic_symbols))
        end = max(end, symbols_end(dynamic_symbols))
    if gnu_debugdata_symbols:
        start = min(start, symbols_start(gnu_debugdata_symbols))
        end = max(end, symbols_end(gnu_debugdata_symbols))

    start &= ~7
    end = (end + 7) & ~7

    # Create a section for the symbols to reference and the corresponding
    # segment for address lookups. It must be SHF_ALLOC and must not be
    # SHT_NOBITS or SHT_NOTE for the file to be loadable.
    size = end - start
    assert size <= 4096, "symbols are too far apart; file would be too large"
    sections = [
        ElfSection(
            name=".data",
            sh_type=SHT.PROGBITS,
            sh_flags=SHF.ALLOC if loadable else 0,
            p_type=PT.LOAD,
            vaddr=start,
            memsz=size,
            data=bytes(size),
        ),
    ]
    symbols = add_shndx(symbols, len(sections))
    dynamic_symbols = add_shndx(dynamic_symbols, len(sections))

    if gnu_debugdata_symbols:
        gds_sections = [
            ElfSection(
                name=".data",
                sh_type=SHT.NOBITS,
                sh_flags=SHF.ALLOC,
                p_type=PT.LOAD,
                vaddr=start,
                memsz=size,
            ),
        ]
        gds_contents = create_elf_file(
            ET.EXEC,
            sections=gds_sections,
            symbols=add_shndx(gnu_debugdata_symbols, len(gds_sections)),
        )
        compressor = lzma.LZMACompressor()
        gds_compressed = compressor.compress(gds_contents) + compressor.flush()
        sections.append(
            ElfSection(
                name=".gnu_debugdata",
                sh_type=SHT.PROGBITS,
                memsz=len(gds_compressed),
                data=gds_compressed,
            )
        )

    if dwarf:
        contents = create_dwarf_file(
            (),
            sections=sections,
            symbols=symbols,
            dynamic_symbols=dynamic_symbols,
        )
    else:
        contents = create_elf_file(
            ET.EXEC,
            sections=sections,
            symbols=symbols,
            dynamic_symbols=dynamic_symbols,
        )

    return contents, start, end


def module_set_elf_symbol_file(module, **kwargs):
    contents, start, end = create_elf_symbol_file(**kwargs)

    with tempfile.NamedTemporaryFile() as f:
        f.write(contents)
        f.flush()

        if module.address_range is None:
            for other_module in module.prog.modules():
                other_address_range = other_module.address_range
                if other_address_range is not None:
                    other_start, other_end = other_address_range
                    assert (
                        end <= other_start or start >= other_end
                    ), f"{module.name} overlaps {other_module.name}"
            module.address_range = (start, end)
        else:
            assert (start, end) == module.address_range

        module.try_file(f.name, force=True)


def program_add_elf_symbol_file(prog, name, **kwargs):
    module = prog.extra_module(name, create=True)
    module_set_elf_symbol_file(module, **kwargs)


def elf_symbol_program(*modules):
    prog = Program()
    for i, symbols in enumerate(modules):
        program_add_elf_symbol_file(prog, f"module{i}", symbols=symbols)
    return prog


class TestElfSymbol(TestCase):
    def assert_symbols_equal_unordered(self, drgn_symbols, symbols):
        self.assertEqual(len(drgn_symbols), len(symbols))
        drgn_symbols = sorted(drgn_symbols, key=lambda x: (x.address, x.name))
        symbols = sorted(symbols, key=lambda x: (x.address, x.name))
        for drgn_symbol, symbol in zip(drgn_symbols, symbols):
            self.assertEqual(drgn_symbol, symbol)

    def test_by_address(self):
        elf_first = ElfSymbol("first", 0xFFFF0000, 0x8, STT.OBJECT, STB.LOCAL)
        elf_second = ElfSymbol("second", 0xFFFF0008, 0x8, STT.OBJECT, STB.LOCAL)
        first = Symbol("first", 0xFFFF0000, 0x8, SymbolBinding.LOCAL, SymbolKind.OBJECT)
        second = Symbol(
            "second", 0xFFFF0008, 0x8, SymbolBinding.LOCAL, SymbolKind.OBJECT
        )

        same_module = ((elf_first, elf_second),)
        different_modules = ((elf_first,), (elf_second,))

        for modules in same_module, different_modules:
            with self.subTest(modules=len(modules)):
                prog = elf_symbol_program(*modules)
                self.assertRaises(LookupError, prog.symbol, 0xFFFEFFFF)
                self.assertEqual(prog.symbols(0xFFFEFFFF), [])
                self.assertEqual(prog.symbol(0xFFFF0000), first)
                self.assert_symbols_equal_unordered(prog.symbols(0xFFFF0000), [first])
                self.assertEqual(prog.symbol(0xFFFF0004), first)
                self.assert_symbols_equal_unordered(prog.symbols(0xFFFF0004), [first])
                self.assertEqual(prog.symbol(0xFFFF0008), second)
                self.assert_symbols_equal_unordered(prog.symbols(0xFFFF0008), [second])
                self.assertEqual(prog.symbol(0xFFFF000C), second)
                self.assert_symbols_equal_unordered(prog.symbols(0xFFFF000C), [second])
                self.assertRaises(LookupError, prog.symbol, 0xFFFF0010)

    def test_by_address_closest(self):
        # If two symbols contain the given address, then the one whose start
        # address is closest to the given address should be preferred
        # (regardless of the binding of either symbol).
        elf_closest = ElfSymbol("closest", 0xFFFF0008, 0x8, STT.OBJECT, STB.WEAK)
        elf_furthest = ElfSymbol("furthest", 0xFFFF0000, 0xC, STT.OBJECT, STB.GLOBAL)
        closest = Symbol(
            "closest", 0xFFFF0008, 0x8, SymbolBinding.WEAK, SymbolKind.OBJECT
        )
        furthest = Symbol(
            "furthest", 0xFFFF0000, 0xC, SymbolBinding.GLOBAL, SymbolKind.OBJECT
        )

        def test(elf_symbols):
            prog = elf_symbol_program(elf_symbols)
            self.assertEqual(prog.symbol(0xFFFF000B), closest)
            self.assert_symbols_equal_unordered(
                prog.symbols(0xFFFF000B), [closest, furthest]
            )

        with self.subTest("closest first"):
            test([elf_closest, elf_furthest])

        with self.subTest("furthest first"):
            test([elf_furthest, elf_closest])

    def test_by_address_closest_end(self):
        # If two symbols contain the given address and have the same start
        # address, then the one whose end address is closest to the given
        # address should be preferred (regardless of the binding of either
        # symbol).
        elf_closest = ElfSymbol("closest", 0xFFFF0000, 0xC, STT.OBJECT, STB.WEAK)
        elf_furthest = ElfSymbol("furthest", 0xFFFF0000, 0x10, STT.OBJECT, STB.GLOBAL)
        closest = Symbol(
            "closest", 0xFFFF0000, 0xC, SymbolBinding.WEAK, SymbolKind.OBJECT
        )
        furthest = Symbol(
            "furthest", 0xFFFF0000, 0x10, SymbolBinding.GLOBAL, SymbolKind.OBJECT
        )

        def test(elf_symbols):
            prog = elf_symbol_program(elf_symbols)
            self.assertEqual(prog.symbol(0xFFFF000B), closest)
            self.assert_symbols_equal_unordered(
                prog.symbols(0xFFFF000B), [closest, furthest]
            )

        with self.subTest("closest first"):
            test([elf_closest, elf_furthest])

        with self.subTest("furthest first"):
            test([elf_furthest, elf_closest])

    def test_by_address_sizeless(self):
        label = ElfSymbol("label", 0xFFFF0008, 0x0, STT.FUNC, STB.LOCAL)
        less = ElfSymbol("less", 0xFFFF0000, 0x4, STT.FUNC, STB.LOCAL)
        greater = ElfSymbol("greater", 0xFFFF0010, 0x4, STT.FUNC, STB.LOCAL)

        expected = Symbol(
            "label", 0xFFFF0008, 0x0, SymbolBinding.LOCAL, SymbolKind.FUNC
        )

        # Test every permutation of every combination of symbols that includes
        # "label".
        for elf_symbols in itertools.chain.from_iterable(
            itertools.permutations((label,) + extra_elf_symbols)
            for r in range(3)
            for extra_elf_symbols in itertools.combinations((less, greater), r)
        ):
            with self.subTest(elf_symbols=[sym.name for sym in elf_symbols]):
                prog = elf_symbol_program(elf_symbols)
                self.assertEqual(prog.symbol(0xFFFF0009), expected)
                self.assertEqual(prog.symbols(0xFFFF0009), [expected])

    def test_by_address_sizeless_subsumed(self):
        label = ElfSymbol("label", 0xFFFF0008, 0x0, STT.FUNC, STB.LOCAL)
        subsume = ElfSymbol("subsume", 0xFFFF0004, 0x8, STT.FUNC, STB.LOCAL)
        less = ElfSymbol("less", 0xFFFF0000, 0x4, STT.FUNC, STB.LOCAL)
        greater = ElfSymbol("greater", 0xFFFF0010, 0x4, STT.FUNC, STB.LOCAL)

        expected = Symbol(
            "subsume", 0xFFFF0004, 0x8, SymbolBinding.LOCAL, SymbolKind.FUNC
        )

        # Test every permutation of every combination of symbols that includes
        # "label" and "subsume".
        for elf_symbols in itertools.chain.from_iterable(
            itertools.permutations((label, subsume) + extra_elf_symbols)
            for r in range(3)
            for extra_elf_symbols in itertools.combinations((less, greater), r)
        ):
            with self.subTest(elf_symbols=[sym.name for sym in elf_symbols]):
                prog = elf_symbol_program(elf_symbols)
                self.assertEqual(prog.symbol(0xFFFF0009), expected)
                self.assertEqual(prog.symbols(0xFFFF0009), [expected])

    def test_by_address_sizeless_wrong_section(self):
        prog = elf_symbol_program(
            (ElfSymbol("label", 0xFFFF0008, 0x0, STT.FUNC, STB.LOCAL),)
        )
        for module in prog.modules():
            start, end = module.address_range
            module.address_range = (start, 0xFFFFFF00)
        self.assertRaises(LookupError, prog.symbol, 0xFFFFFE00)

    def test_by_address_binding_precedence(self):
        precedence = (
            (STB.GLOBAL, STB.GNU_UNIQUE),
            (STB.WEAK,),
            (STB.LOCAL, STB.HIPROC),
        )

        def assert_find_higher(*modules, both):
            prog = elf_symbol_program(*modules)
            self.assertEqual(prog.symbol(0xFFFF0000).name, "foo")
            # Test that symbols() finds both if expected or either one if not.
            if both:
                self.assertCountEqual(
                    [sym.name for sym in prog.symbols(0xFFFF0000)], ["foo", "bar"]
                )
            else:
                self.assertIn(
                    [sym.name for sym in prog.symbols(0xFFFF0000)], (["foo"], ["bar"])
                )

        for size in (8, 0):
            with self.subTest(size=size):
                for i in range(len(precedence) - 1):
                    for higher_binding in precedence[i]:
                        for j in range(i + 1, len(precedence)):
                            for lower_binding in precedence[j]:
                                with self.subTest(
                                    higher=higher_binding, lower=lower_binding
                                ):
                                    higher = ElfSymbol(
                                        "foo",
                                        0xFFFF0000,
                                        size,
                                        STT.OBJECT,
                                        higher_binding,
                                    )
                                    lower = ElfSymbol(
                                        "bar",
                                        0xFFFF0000,
                                        size,
                                        STT.OBJECT,
                                        lower_binding,
                                    )
                                    # Local symbols must be before global symbols.
                                    if lower_binding not in precedence[-1]:
                                        with self.subTest("higher before lower"):
                                            assert_find_higher(
                                                (higher, lower), both=size > 0
                                            )
                                    with self.subTest("lower before higher"):
                                        assert_find_higher(
                                            (lower, higher), both=size > 0
                                        )

    def test_by_name(self):
        elf_first = ElfSymbol("first", 0xFFFF0000, 0x8, STT.OBJECT, STB.GLOBAL)
        elf_second = ElfSymbol("second", 0xFFFF0008, 0x8, STT.OBJECT, STB.GLOBAL)
        first = Symbol(
            "first", 0xFFFF0000, 0x8, SymbolBinding.GLOBAL, SymbolKind.OBJECT
        )
        second = Symbol(
            "second", 0xFFFF0008, 0x8, SymbolBinding.GLOBAL, SymbolKind.OBJECT
        )

        same_module = ((elf_first, elf_second),)
        different_modules = ((elf_first,), (elf_second,))

        for modules in same_module, different_modules:
            with self.subTest(modules=len(modules)):
                prog = elf_symbol_program(*modules)
                self.assertEqual(prog.symbol("first"), first)
                self.assertEqual(prog.symbol("second"), second)
                self.assertRaises(LookupError, prog.symbol, "third")

                self.assert_symbols_equal_unordered(prog.symbols("first"), [first])
                self.assert_symbols_equal_unordered(prog.symbols("second"), [second])
                self.assertEqual(prog.symbols("third"), [])

    def test_by_name_binding_precedence(self):
        precedence = (
            (STB.GLOBAL, STB.GNU_UNIQUE),
            (STB.WEAK,),
            (STB.LOCAL, STB.HIPROC),
        )

        expected = 0xFFFF0008
        other = expected - 0x8

        def assert_find_higher(*modules):
            prog = elf_symbol_program(*modules)
            self.assertEqual(prog.symbol("foo").address, expected)
            # assert symbols() always finds both
            self.assertCountEqual(
                [sym.address for sym in prog.symbols("foo")], [expected, other]
            )

        for i in range(len(precedence) - 1):
            for higher_binding in precedence[i]:
                for j in range(i + 1, len(precedence)):
                    for lower_binding in precedence[j]:
                        with self.subTest(higher=higher_binding, lower=lower_binding):
                            higher = ElfSymbol(
                                "foo", expected, 0x8, STT.OBJECT, higher_binding
                            )
                            lower = ElfSymbol(
                                "foo", other, 0x8, STT.OBJECT, lower_binding
                            )
                            # Local symbols must be before global symbols.
                            if lower_binding not in precedence[-1]:
                                with self.subTest("same module, higher before lower"):
                                    assert_find_higher((higher, lower))
                            with self.subTest("same module, lower before higher"):
                                assert_find_higher((lower, higher))
                            with self.subTest("different modules, higher before lower"):
                                assert_find_higher((higher,), (lower,))
                            with self.subTest("different modules, lower before higher"):
                                assert_find_higher((lower,), (higher,))

    def test_binding(self):
        for by in "name", "address":
            for elf_binding, drgn_binding in (
                (STB.LOCAL, SymbolBinding.LOCAL),
                (STB.GLOBAL, SymbolBinding.GLOBAL),
                (STB.WEAK, SymbolBinding.WEAK),
                (STB.GNU_UNIQUE, SymbolBinding.UNIQUE),
                (STB.HIPROC, SymbolBinding.UNKNOWN),
            ):
                with self.subTest(by=by, binding=elf_binding):
                    prog = elf_symbol_program(
                        (ElfSymbol("foo", 0xFFFF0000, 1, STT.OBJECT, elf_binding),)
                    )
                    self.assertEqual(
                        prog.symbol("foo" if by == "name" else 0xFFFF0000).binding,
                        drgn_binding,
                    )
                    if by == "name":
                        symbols = prog.symbols("foo")
                        self.assertEqual(len(symbols), 1)
                        self.assertEqual(symbols[0].binding, drgn_binding)

    def test_kind(self):
        for elf_type, drgn_kind in (
            (STT.NOTYPE, SymbolKind.UNKNOWN),
            (STT.OBJECT, SymbolKind.OBJECT),
            (STT.FUNC, SymbolKind.FUNC),
            (STT.SECTION, SymbolKind.SECTION),
            (STT.FILE, SymbolKind.FILE),
            (STT.COMMON, SymbolKind.COMMON),
            (STT.TLS, SymbolKind.TLS),
            (STT.GNU_IFUNC, SymbolKind.IFUNC),
        ):
            with self.subTest(type=elf_type):
                prog = elf_symbol_program(
                    (ElfSymbol("foo", 0xFFFF0000, 1, elf_type, STB.GLOBAL),)
                )
                symbol = Symbol("foo", 0xFFFF0000, 1, SymbolBinding.GLOBAL, drgn_kind)
                self.assertEqual(prog.symbol("foo"), symbol)
                symbols = prog.symbols("foo")
                self.assert_symbols_equal_unordered(symbols, [symbol])

    def test_all_symbols(self):
        elf_syms = (
            (
                ElfSymbol("two", 0xFFFF0012, 1, STT.OBJECT, STB.LOCAL),
                ElfSymbol("three", 0xFFFF0013, 1, STT.OBJECT, STB.LOCAL),
                ElfSymbol("one", 0xFFFF0011, 1, STT.OBJECT, STB.GLOBAL),
            ),
            (
                ElfSymbol("three", 0xFFFF0023, 1, STT.OBJECT, STB.LOCAL),
                ElfSymbol("two", 0xFFFF0022, 1, STT.OBJECT, STB.GLOBAL),
            ),
            (ElfSymbol("three", 0xFFFF0033, 1, STT.OBJECT, STB.GLOBAL),),
        )
        kind = SymbolKind.OBJECT
        syms = [
            Symbol("two", 0xFFFF0012, 1, SymbolBinding.LOCAL, kind),
            Symbol("three", 0xFFFF0013, 1, SymbolBinding.LOCAL, kind),
            Symbol("one", 0xFFFF0011, 1, SymbolBinding.GLOBAL, kind),
            Symbol("three", 0xFFFF0023, 1, SymbolBinding.LOCAL, kind),
            Symbol("two", 0xFFFF0022, 1, SymbolBinding.GLOBAL, kind),
            Symbol("three", 0xFFFF0033, 1, SymbolBinding.GLOBAL, kind),
        ]
        prog = elf_symbol_program(*elf_syms)
        self.assert_symbols_equal_unordered(prog.symbols(), syms)

    def test_dynsym(self):
        prog = Program()
        program_add_elf_symbol_file(
            prog,
            "module0",
            dynamic_symbols=[
                ElfSymbol("sym", 0xFFFF0000, 0x8, STT.OBJECT, STB.LOCAL),
            ],
        )

        sym = Symbol("sym", 0xFFFF0000, 0x8, SymbolBinding.LOCAL, SymbolKind.OBJECT)
        self.assertEqual(prog.symbol("sym"), sym)
        self.assertEqual(prog.symbol(0xFFFF0004), sym)

    def test_ignore_dynsym_same_file(self):
        # Test that .dynsym is ignored in a file with both .symtab and .dynsym.
        prog = Program()
        program_add_elf_symbol_file(
            prog,
            "module0",
            # Normally .symtab is a superset of .dynsym, but to test that we
            # ignore .dynsym, make them distinct.
            symbols=[
                ElfSymbol("full", 0xFFFF0000, 0x8, STT.OBJECT, STB.LOCAL),
            ],
            dynamic_symbols=[
                ElfSymbol("partial", 0xFFFF0000, 0x8, STT.OBJECT, STB.LOCAL),
            ],
        )

        self.assertRaises(LookupError, prog.symbol, "partial")

        full = Symbol("full", 0xFFFF0000, 0x8, SymbolBinding.LOCAL, SymbolKind.OBJECT)
        self.assertEqual(prog.symbol("full"), full)
        self.assertEqual(prog.symbol(0xFFFF0004), full)

    def test_ignore_dynsym_separate_files(self):
        # Same as test_ignore_dynsym_same_file(), except .symtab and .dynsym
        # are in different files.
        prog = Program()
        program_add_elf_symbol_file(
            prog,
            "module0",
            dynamic_symbols=[
                ElfSymbol("partial", 0xFFFF0000, 0x8, STT.OBJECT, STB.LOCAL),
            ],
        )
        program_add_elf_symbol_file(
            prog,
            "module0",
            symbols=[
                ElfSymbol("full", 0xFFFF0000, 0x8, STT.OBJECT, STB.LOCAL),
            ],
            dwarf=True,
        )

        self.assertRaises(LookupError, prog.symbol, "partial")

        full = Symbol("full", 0xFFFF0000, 0x8, SymbolBinding.LOCAL, SymbolKind.OBJECT)
        self.assertEqual(prog.symbol("full"), full)
        self.assertEqual(prog.symbol(0xFFFF0004), full)

    def test_override_dynsym(self):
        # Same as test_ignore_dynsym_separate_files(), except we do a lookup in
        # .dynsym before we have .symtab.
        prog = Program()
        program_add_elf_symbol_file(
            prog,
            "module0",
            dynamic_symbols=[
                ElfSymbol("partial", 0xFFFF0000, 0x8, STT.OBJECT, STB.LOCAL),
            ],
        )

        partial = Symbol(
            "partial", 0xFFFF0000, 0x8, SymbolBinding.LOCAL, SymbolKind.OBJECT
        )
        self.assertEqual(prog.symbol("partial"), partial)
        self.assertEqual(prog.symbol(0xFFFF0004), partial)

        program_add_elf_symbol_file(
            prog,
            "module0",
            symbols=[
                ElfSymbol("full", 0xFFFF0000, 0x8, STT.OBJECT, STB.LOCAL),
            ],
            dwarf=True,
        )

        self.assertRaises(LookupError, prog.symbol, "partial")

        full = Symbol("full", 0xFFFF0000, 0x8, SymbolBinding.LOCAL, SymbolKind.OBJECT)
        self.assertEqual(prog.symbol("full"), full)
        self.assertEqual(prog.symbol(0xFFFF0004), full)


@unittest.skipUnless(drgn._with_lzma, "built without lzma support")
class TestGnuDebugdata(TestCase):

    def assert_all_symbols_found_by_name(self, prog, symbols):
        for symbol in symbols:
            self.assertEqual(prog.symbol(symbol.name), symbol)

    def assert_all_symbols_found_by_address(self, prog, symbols):
        for symbol in symbols:
            self.assertEqual(prog.symbol(symbol.address), symbol)
            self.assertEqual(prog.symbol(symbol.address + symbol.size - 1), symbol)

    def assert_all_symbols_returned_by_lookup(self, prog, symbols):
        def sort_key(sym):
            return (sym.address, sym.name)

        expected = sorted(symbols, key=sort_key)
        actual = prog.symbols()
        actual.sort(key=sort_key)
        self.assertEqual(expected, actual)

    def test_gnu_debugdata_and_dynamic_lookup(self):
        gnu_symbols = [
            ElfSymbol("first", 0xFFFF0000, 0x8, STT.FUNC, STB.LOCAL),
            ElfSymbol("second", 0xFFFF0018, 0x8, STT.FUNC, STB.LOCAL),
        ]
        dynamic_symbols = [
            ElfSymbol("third", 0xFFFF0010, 0x8, STT.FUNC, STB.LOCAL),
            ElfSymbol("fourth", 0xFFFF0008, 0x8, STT.FUNC, STB.LOCAL),
        ]
        prog = Program()
        program_add_elf_symbol_file(
            prog,
            "module0",
            dynamic_symbols=dynamic_symbols,
            gnu_debugdata_symbols=gnu_symbols,
        )
        drgn_symbols = [
            Symbol("first", 0xFFFF0000, 0x8, SymbolBinding.LOCAL, SymbolKind.FUNC),
            Symbol("second", 0xFFFF0018, 0x8, SymbolBinding.LOCAL, SymbolKind.FUNC),
            Symbol("third", 0xFFFF0010, 0x8, SymbolBinding.LOCAL, SymbolKind.FUNC),
            Symbol("fourth", 0xFFFF0008, 0x8, SymbolBinding.LOCAL, SymbolKind.FUNC),
        ]
        self.assert_all_symbols_found_by_name(prog, drgn_symbols)
        self.assert_all_symbols_found_by_address(prog, drgn_symbols)
        self.assert_all_symbols_returned_by_lookup(prog, drgn_symbols)

    def test_sizeless_symbols_gnu_debugdata(self):
        gnu_symbols = [
            ElfSymbol("zero", 0xFFFF0000, 0x0, STT.FUNC, STB.LOCAL),
            ElfSymbol("two", 0xFFFF0002, 0x4, STT.FUNC, STB.LOCAL),
            ElfSymbol("ten", 0xFFFF000A, 0x0, STT.FUNC, STB.LOCAL),
        ]
        dynamic_symbols = [
            ElfSymbol("four", 0xFFFF0004, 0x0, STT.FUNC, STB.LOCAL),
            ElfSymbol("eight", 0xFFFF0008, 0x0, STT.FUNC, STB.LOCAL),
        ]
        drgn_symbols = {
            s.name: s
            for s in (
                Symbol("zero", 0xFFFF0000, 0x0, SymbolBinding.LOCAL, SymbolKind.FUNC),
                Symbol("two", 0xFFFF0002, 0x4, SymbolBinding.LOCAL, SymbolKind.FUNC),
                Symbol("four", 0xFFFF0004, 0x0, SymbolBinding.LOCAL, SymbolKind.FUNC),
                Symbol("eight", 0xFFFF0008, 0x0, SymbolBinding.LOCAL, SymbolKind.FUNC),
                Symbol("ten", 0xFFFF000A, 0x0, SymbolBinding.LOCAL, SymbolKind.FUNC),
            )
        }

        for swap in (False, True):
            prog = Program()
            program_add_elf_symbol_file(
                prog,
                "module0",
                dynamic_symbols=gnu_symbols if swap else dynamic_symbols,
                gnu_debugdata_symbols=dynamic_symbols if swap else gnu_symbols,
            )

            self.assert_all_symbols_found_by_name(prog, drgn_symbols.values())
            self.assert_all_symbols_returned_by_lookup(prog, drgn_symbols.values())

            # Address 9 has a best match in .dynsym, despite other sizeless matches
            # in .gnu_debugdata.
            self.assertEqual(drgn_symbols["eight"], prog.symbol(0xFFFF0009))

            # Address 5 is conained by symbol "two" in .gnu_debugdata, despite
            # "four" being a sizeless match in .dynsym.
            self.assertEqual(drgn_symbols["two"], prog.symbol(0xFFFF0005))

            # Address 11 has a best sizeless match of "ten" in .gnu_debugdata,
            # despite having a sizeless match of "eight" in .dynsym.
            self.assertEqual(drgn_symbols["ten"], prog.symbol(0xFFFF000B))

    def test_file_preferences(self):
        # We need to be careful to make the address range the same for both
        # files: so the minimum and maximum address for gnu + dynamic must be
        # the same as for symtab.
        # Normally a debug file would contain the same symbols as the loaded
        # file, plus more. For testing, give them different names to
        # distinguish.
        loaded = [
            ElfSymbol("loaded_lo", 0xFFFF0000, 0x4, STT.FUNC, STB.LOCAL),
            ElfSymbol("loaded_hi", 0xFFFF0004, 0x4, STT.FUNC, STB.LOCAL),
        ]
        debug = [
            ElfSymbol("symtab_lo", 0xFFFF0000, 0x4, STT.OBJECT, STB.LOCAL),
            ElfSymbol("symtab_hi", 0xFFFF0004, 0x4, STT.OBJECT, STB.LOCAL),
        ]
        empty = [ElfSymbol("", 0xFFFF0000, 0, 0, 0, 0, 0)]
        loaded_file_symbols = [
            Symbol("loaded_lo", 0xFFFF0000, 0x4, SymbolBinding.LOCAL, SymbolKind.FUNC),
            Symbol("loaded_hi", 0xFFFF0004, 0x4, SymbolBinding.LOCAL, SymbolKind.FUNC),
        ]
        debug_file_symbols = [
            Symbol(
                "symtab_lo", 0xFFFF0000, 0x4, SymbolBinding.LOCAL, SymbolKind.OBJECT
            ),
            Symbol(
                "symtab_hi", 0xFFFF0004, 0x4, SymbolBinding.LOCAL, SymbolKind.OBJECT
            ),
        ]
        file_choices = {
            "loaded": (
                {"gnu_debugdata_symbols": loaded[:1], "dynamic_symbols": loaded[1:]},
                loaded_file_symbols,
            ),
            "loaded_dyn": (
                {"dynamic_symbols": loaded},
                loaded_file_symbols,
            ),
            "loaded_gnu": (
                {"gnu_debugdata_symbols": loaded},
                loaded_file_symbols,
            ),
            "loaded_gnu_dynempty": (
                {"gnu_debugdata_symbols": loaded, "dynamic_symbols": empty},
                loaded_file_symbols,
            ),
            "debug": (
                {"symbols": debug, "dwarf": True, "loadable": False},
                debug_file_symbols,
            ),
            "debug_dyn": (
                {"dynamic_symbols": debug, "dwarf": True, "loadable": False},
                debug_file_symbols,
            ),
        }

        # First file, second file, whether or not the symtab should be replaced.
        # Combining the symbol table is possible in a corner case (.dynsym from
        # the debug file, plus .gnu_debugdata from the loaded, if the loaded
        # file has no .dynsym of its own). This really ought not to happen in
        # practice, but it's worth ensuring that it's handled safely.
        cases = [
            ("loaded", "debug", "replace"),
            ("loaded_dyn", "debug", "replace"),
            ("loaded_gnu", "debug", "replace"),
            ("loaded_gnu_dynempty", "debug", "replace"),
            ("debug", "loaded", None),
            ("debug", "loaded_dyn", None),
            ("debug", "loaded_gnu", None),
            ("debug", "loaded_gnu_dynempty", None),
            ("loaded", "debug_dyn", None),
            ("loaded_dyn", "debug_dyn", None),
            ("loaded_gnu", "debug_dyn", "combine"),
            ("loaded_gnu_dynempty", "debug_dyn", None),
            # We will replace a .dynsym with another .dynsym only if the file
            # also has a .gnu_debugdata
            ("debug_dyn", "loaded", "replace"),
            ("debug_dyn", "loaded_dyn", None),
            ("debug_dyn", "loaded_gnu", "combine"),
            ("debug_dyn", "loaded_gnu_dynempty", "replace"),
        ]

        for first, second, action in cases:
            with self.subTest(f"{first}, {second}"):
                prog = Program()
                module = prog.extra_module("module0", create=True)
                module_set_elf_symbol_file(module, **file_choices[first][0])
                expected = file_choices[first][1]
                self.assert_all_symbols_found_by_name(prog, expected)
                self.assert_all_symbols_found_by_address(prog, expected)
                self.assert_all_symbols_returned_by_lookup(prog, expected)

                module_set_elf_symbol_file(module, **file_choices[second][0])
                if action == "replace":
                    expected = file_choices[second][1]
                elif action == "combine":
                    expected = expected + file_choices[second][1]
                self.assert_all_symbols_found_by_name(prog, expected)
                # We end up with overlapping symbols when tables get combined.
                # Don't bother checking address lookup there.
                if action != "combine":
                    self.assert_all_symbols_found_by_address(prog, expected)
                self.assert_all_symbols_returned_by_lookup(prog, expected)


class TestSymbolFinder(TestCase):
    TEST_SYMS = [
        Symbol("one", 0xFFFF1000, 16, SymbolBinding.LOCAL, SymbolKind.FUNC),
        Symbol("two", 0xFFFF2000, 16, SymbolBinding.GLOBAL, SymbolKind.FUNC),
        Symbol("three", 0xFFFF2008, 8, SymbolBinding.GLOBAL, SymbolKind.FUNC),
    ]

    def finder(self, prog, arg_name, arg_address, arg_one):
        self.called = True
        res = []
        self.assertEqual(self.expected_name, arg_name)
        self.assertEqual(self.expected_address, arg_address)
        self.assertEqual(self.expected_one, arg_one)
        for sym in self.TEST_SYMS:
            if arg_name and sym.name == arg_name:
                res.append(sym)
            elif arg_address and sym.address <= arg_address < sym.address + sym.size:
                res.append(sym)
            elif not arg_name and not arg_address:
                res.append(sym)

        # This symbol finder intentionally has a bug: it does not respect the
        # "arg_one" flag: it may return multiple symbols even when "arg_one" is
        # true.
        return res

    def setUp(self):
        self.prog = Program()
        self.prog.register_symbol_finder("test", self.finder, enable_index=0)
        self.called = False

    def expect_args(self, name, address, one):
        self.expected_name = name
        self.expected_address = address
        self.expected_one = one

    def test_args_single_string(self):
        self.expect_args("search_symbol", None, True)
        with self.assertRaises(LookupError):
            self.prog.symbol("search_symbol")
        self.assertTrue(self.called)

    def test_args_single_int(self):
        self.expect_args(None, 0xFF00, True)
        with self.assertRaises(LookupError):
            self.prog.symbol(0xFF00)
        self.assertTrue(self.called)

    def test_args_single_with_many_results(self):
        self.expect_args(None, 0xFFFF2008, True)
        with self.assertRaises(ValueError):
            self.prog.symbol(0xFFFF2008)
        self.assertTrue(self.called)

    def test_single_with_result(self):
        self.expect_args("one", None, True)
        self.assertEqual(self.prog.symbol("one"), self.TEST_SYMS[0])
        self.assertTrue(self.called)

    def test_args_many_string(self):
        self.expect_args("search_symbol", None, False)
        self.assertEqual(self.prog.symbols("search_symbol"), [])
        self.assertTrue(self.called)

    def test_args_many_int(self):
        self.expect_args(None, 0xFF00, False)
        self.assertEqual(self.prog.symbols(0xFF00), [])
        self.assertTrue(self.called)

    def test_many_with_result(self):
        self.expect_args(None, 0xFFFF2004, False)
        self.assertEqual(self.prog.symbols(0xFFFF2004), [self.TEST_SYMS[1]])
        self.assertTrue(self.called)

    def test_many_without_filter(self):
        self.expect_args(None, None, False)
        self.assertEqual(self.prog.symbols(), self.TEST_SYMS)
        self.assertTrue(self.called)


class TestSymbolIndex(TestCase):
    # Symbols are listed here in order of address, but are shuffled below
    AA = Symbol("AA", 10, 5, SymbolBinding.GLOBAL, SymbolKind.OBJECT)
    BB = Symbol("BB", 12, 1, SymbolBinding.GLOBAL, SymbolKind.OBJECT)
    CC = Symbol("CC", 13, 8, SymbolBinding.GLOBAL, SymbolKind.OBJECT)
    DD = Symbol("DD", 28, 5, SymbolBinding.GLOBAL, SymbolKind.OBJECT)
    EE = Symbol("EE", 34, 1, SymbolBinding.GLOBAL, SymbolKind.OBJECT)
    FF = Symbol("FF", 34, 10, SymbolBinding.GLOBAL, SymbolKind.OBJECT)
    GG = Symbol("GG", 34, 2, SymbolBinding.GLOBAL, SymbolKind.OBJECT)
    BB2 = Symbol("BB", 36, 3, SymbolBinding.GLOBAL, SymbolKind.OBJECT)

    TEST_SYMS = [GG, BB, AA, BB2, CC, FF, DD, EE]

    def setUp(self):
        # This class tests both the SymbolIndex callable interface, and the
        # Symbol Finder API. While this seems like it duplicates code, it's
        # necessary to test both since they exercise different code paths: the
        # Symbol Finder API uses a more efficient fast path.
        self.finder = SymbolIndex(self.TEST_SYMS)
        self.prog = Program()
        self.prog.register_symbol_finder("test", self.finder, enable_index=0)

    def test_name_single(self):
        for sym in self.TEST_SYMS:
            if sym.name != "BB":
                self.assertEqual([sym], self.finder(self.prog, sym.name, None, True))
                self.assertEqual(sym, self.prog.symbol(sym.name))
                self.assertEqual([sym], self.finder(self.prog, sym.name, None, False))
                self.assertEqual([sym], self.prog.symbols(sym.name))

    def test_name_multiple(self):
        multi_result = self.finder(self.prog, "BB", None, False)
        self.assertEqual(2, len(multi_result))
        self.assertIn(self.BB, multi_result)
        self.assertIn(self.BB2, multi_result)

        multi_result = self.prog.symbols("BB")
        self.assertEqual(2, len(multi_result))
        self.assertIn(self.BB, multi_result)
        self.assertIn(self.BB2, multi_result)

        single_result = self.finder(self.prog, "BB", None, True)
        self.assertIn(single_result[0], (self.BB, self.BB2))

        single_result = self.prog.symbol("BB")
        self.assertIn(single_result, (self.BB, self.BB2))

    def test_addr(self):
        cases = {
            9: [],
            10: [self.AA],
            12: [self.AA, self.BB],
            13: [self.AA, self.CC],
            15: [self.CC],
            25: [],
            28: [self.DD],
            30: [self.DD],
            34: [self.EE, self.FF, self.GG],
            35: [self.FF, self.GG],
            36: [self.FF, self.BB2],
            43: [self.FF],
            44: [],
        }
        for address, expected in cases.items():
            # first, lookup by address alone and ensure we get all correct
            # candidates:
            multi_result = self.finder(self.prog, None, address, False)
            self.assertEqual(len(expected), len(multi_result))
            self.assertTrue(all(e in multi_result for e in expected))
            multi_result = self.prog.symbols(address)
            self.assertEqual(len(expected), len(multi_result))
            self.assertTrue(all(e in multi_result for e in expected))

            # next, ensure that the single lookup works as expected:
            if expected:
                single_result = self.finder(self.prog, None, address, True)
                self.assertEqual(1, len(single_result))
                self.assertIn(single_result[0], expected)
                single_result = self.prog.symbol(address)
                self.assertIn(single_result, expected)

            # Now, test that adding a name filter correctly filters:
            # This cannot be tested with the Program.symbol() API since only
            # one filter is allowed there.
            for sym in expected:
                self.assertEqual([sym], self.finder(self.prog, sym.name, address, True))
                self.assertEqual(
                    [sym], self.finder(self.prog, sym.name, address, False)
                )

            self.assertEqual([], self.finder(None, "MISSING", address, True))
            self.assertEqual([], self.finder(None, "MISSING", address, False))

    def test_all(self):
        result = self.finder(self.prog, None, None, True)
        self.assertEqual(1, len(result))
        self.assertIn(result[0], self.TEST_SYMS)
        result = self.finder(self.prog, None, None, False)
        self.assertEqual(len(self.TEST_SYMS), len(result))
        for sym in self.TEST_SYMS:
            self.assertIn(sym, result)
        result = self.prog.symbols()
        self.assertEqual(len(self.TEST_SYMS), len(result))
        for sym in self.TEST_SYMS:
            self.assertIn(sym, result)

    def test_empty_index(self):
        index = SymbolIndex([])
        # Check all the possible query patterns to ensure they can safely handle
        # an empty list.
        self.assertEqual([], index(self.prog, "name search", None, True))
        self.assertEqual([], index(self.prog, "name search", None, False))
        self.assertEqual([], index(self.prog, None, 0xFFFF, True))
        self.assertEqual([], index(self.prog, None, 0xFFFF, False))
        self.assertEqual([], index(self.prog, "name search", 0xFFFF, True))
        self.assertEqual([], index(self.prog, "name search", 0xFFFF, False))



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

from drgn.helpers.linux.sbitmap import sbitmap_for_each_set
from tests.linux_kernel import LinuxKernelTestCase, skip_unless_have_test_kmod


@skip_unless_have_test_kmod
class TestSbitmap(LinuxKernelTestCase):
    def test_sbitmap_for_each_set(self):
        self.assertEqual(
            list(sbitmap_for_each_set(self.prog["drgn_test_sbitmap"].address_of_())),
            [13, 23, 24, 99],
        )
